"""
Default Units:

- Distance: Angstrom, A
- Time: Millisecond, ms
- Magnetic Field: Gaussian, G = 1e-4 Tesla
- Gyromagnetic Ratio: rad / ms / Gauss
- Quadrupole moment: barn
- Couplings: kHz

"""

import warnings

import numpy as np
from pycce.io.xyz import read_xyz

from .bath.array import BathArray, SpinDict, broadcast_array
from .bath.cube import Cube
from .center import CenterArray
from .constants import ELECTRON_GYRO
from .find_clusters import generate_clusters
from .run.cce import CCE
from .run.corr import CCENoise, gCCENoise
from .run.gcce import gCCE
from .run.pulses import Sequence
from .utilities import _add_args

_returns = r"""

            Returns:
    
                ndarray: Computed property."""

_args = r"""
            magnetic_field (ndarray with shape (3,) or callable): Magnetic field vector of form (Bx, By, Bz)
                or callable with signature ``magnetic_field(pos)``, where ``pos`` is an array with shape (3,) with the
                position of the spin. 
                
                Default is **None**. Overrides ``Simulator.magnetic_field`` if provided.

            pulses (list or int or Sequence): Number of pulses in CPMG sequence.

                **OR**

                Sequence of the instantaneous ideal control pulses.
                It can be provided as an instance of ``Sequence`` class or a list with ``Pulse`` objects.
                (See documentation for pycce.Sequence).
                
                ``pulses`` can be provided as a list with tuples or dictionaries,
                each tuple or dictionary is used to initialize ``Pulse`` class instance.

                For example, for only central spin pulses the ``pulses`` argument
                can be provided as a list of tuples, containing:

                1. axis the rotation is about;
                2. angle of rotation;
                3. (optional) Time before the pulse. Can be as fixed, as well as varied.
                   If varied, it should be provided as an array with the same
                   length as ``timespace``.

                E.g. for Hahn-Echo the ``pulses`` can be defined as ``[('x', np.pi)]`` or 
                ``[('x', np.pi, timespace / 2)]``.

                .. note::
                
                    If delay is not provided in **all** pulses, assumes even delay of CPMG sequence. 
                    If only **some** delays are provided, assumes ``delay = 0``  in the pulses without delay.

                    Then total experiment is assumed to be:

                        tau -- pulse -- 2tau -- pulse -- ... -- 2tau -- pulse -- tau

                    Where tau is the delay between pulses.

                    The sum of delays at each time point should be less or equal to the total time of the experiment
                    at the same time point, provided in ``timespace`` argument. 

                .. warning::
    
                    In conventional CCE calculations, only :math:`pi` pulses
                    on the central spin are allowed. 

                In the calculations of noise autocorrelation this parameter is ignored.

                Default is **None**. Overrides``Simulator.pulses`` if provided.
            
            i (int or ndarray with shape (2s+1, ) or callable): Used in gCCE calculations.
                Along with ``j`` parameter indicates which density matrix element to compute with gCCE as:
                
                .. math::
                
                    L=\bra{i}\hat \rho \ket{j}
                
                By default is equal to :math:`R\ket{0}` state of the ``.center``
                where :math:`R` is a product of all rotations applied in the pulse sequence.
                Can be set as a vector in :math:`S_z` basis, the index of the central spin Hamiltonian
                eigenstate, or as a callable with call signature ``i(dm)``, where ``dm`` is a density matrix of the 
                central spin. If callable, ``j`` parameter is ignored.

            j (int or ndarray with shape (2s+1, ) or callable): Used in gCCE calculations.
                Along with ``i`` parameter indicates which density matrix element to compute.
                
                By default is equal to :math:`R\ket{1}` state of the ``.center``
                where :math:`R` is a product of all rotations applied in the pulse sequence.
                Can be set as a vector in :math:`S_z` basis, the index of the central spin Hamiltonian
                eigenstate, or as a callable with call signature ``j(dm)``, where ``dm`` is a density matrix of the 
                central spin. If callable, ``i`` parameter is ignored.
    
            as_delay (bool): True if time points are delay between pulses (for equispaced pulses),
                False if time points are total time. Ignored in gCCE if ``pulses`` contains the time fractions.
                Conventional CCE calculations do not support custom time fractions.

                Default is **False**.

            interlaced (bool): True if use hybrid CCE approach - for each cluster 
                sample over states of the supercluster.

                Default is **False**.

            state (ndarray with shape (2s+1,)):
                Initial state of the central spin, used in gCCE and noise autocorrelation calculations.
                
                Defaults to :math:`\frac{1}{N}(\ket{0} + \ket{1})` if not set.
    
            bath_state (array-like):
                List of bath spin states. If ``len(shape) == 1``, contains
                :math:`I_z` projections of :math:`I_z` eigenstates.
                Otherwise, contains array of initial density matrices of bath spins.
    
                Default is **None**. If not set, the code assumes completely random spin bath
                (density matrix of each nuclear spin is proportional to identity, :math:`\hat {\mathbb{1}}/N`).

            nbstates (int): Number or random bath states to sample over.

                If provided, sampling of random states is carried and ``bath_states`` values are
                ignored.
                    
                Default is 0.

            seed (int): Seed for random number generator, used in random bath states sampling.
                
                Default is **None**.

            masked (bool):
                True if mask numerically unstable points (with coherence > 1)
                in the averaging over bath states. 
                
                .. note::
                
                    It is up to user to check whether the possible instability is due to numerical error
                    or unphysical assumptions of the calculations.

                Default is **True** for coherence calculations, **False** for noise calculations.

            parallel_states (bool):
                True if to use MPI to parallelize the calculations of density matrix equally over
                present mpi processes for random bath state sampling calculations.
                
                Compared to ``parallel`` keyword,
                when this argument is True each process is given a fraction of random bath states.
                This makes the implementation faster. Works best when the
                number of bath states is divisible by the number of processes, ``nbstates % size == 0``.
    
                Default is **False**.

            second_order (bool):
                True if add second order perturbation theory correction
                to the cluster Hamiltonian in conventional CCE.
                Relevant only for conventional CCE calculations.
                
                If set to True sets the qubit states as eigenstates of central spin Hamiltonian
                from the following procedure. If qubit states are provided as vectors in :math:`S_z` basis,
                for each qubit state compute the fidelity of the qubit state and
                all eigenstates of the central spin and chose the one with fidelity higher than ``level_confidence``.
                If such state is not found, raises an error.
    
                .. warning::
    
                    Second order corrections are not implemented as mean field.
                    
                    I.e., setting ``second_order=True``
                    and ``nbstates != 0`` leads to the calculation, when mean field effect is accounted only from
                    dipolar interactions within the bath.
    
                Default is **False**.
    
            level_confidence (float): Maximum fidelity of the qubit state to be considered eigenstate of the
                central spin Hamiltonian. 
                
                Default is 0.95.
    
            direct (bool):
                True if use direct approach (requires way more memory but might be more numerically stable).
                False if use memory efficient approach.
                
                Default is **False**.
    
            parallel (bool):
                True if parallelize calculation of cluster contributions over different mpi processes.

                Default is **False**."""


# TODO unit conversion
class Simulator:
    r"""
    The main class for CCE calculations.

    The typical usage includes:

        1. Read array of the bath spins. This is done with ``Simulator.read_bath`` method which accepts either
           reading from .xyz file or from the ``BathArray`` instance with defined positions and names of the bath spins.
           In the process, the subset of the array within the distance of ``r_dipole`` from the central spin is taken
           and for this subset the Hyperfine couplings can be generated.

           If no ``hyperfine`` keyword is provided and there are some hyperfine couplings already,
           then no changes are done to the hyperfine tensors.
           If ``hyperfine='pd'``, the hyperfine couplings are computed assuming point dipole approximation. For all
           accepted arguments, see ``Simulator.read_bath``.

        2. Generate set of clusters with ``Simulator.generate_clusters``,
           determined by the maximum connectivity radius ``r_dipole``
           and the maximum size of the cluster ``order``.

        3. Compute the desired property with ``Simulator.compute`` method.

    .. note::

        Directly setting up the attribute values will rerun ``Simulator.read_bath``
        and/or ``Simulator.generate_clusters`` to reflect updated value of the given attribute.

        E.g. If ``Simulator.r_bath`` is set to some new value after initialization,
        then ``Simulator.read_bath`` and ``Simulator.generate_clusters`` are called with the increased bath.

    First two steps are usually done during the initialization of the ``Simulator`` object by providing the necessary
    arguments.

    Notes:

        Depending on the number of provided arguments, in the initialization process will call the following methods
        to setup the calculation engine.

        - If ``bath`` is provided, ``Simulator.read_bath`` is called with additional keywords in ``**bath_kw``.
        - If both ``r_dipole`` and ``order`` are provided and ``bath`` is not None,
          the ``Simulator.generate_clusters`` is called.

        See the corresponding method documentation for details.

    Examples::

        >>> atoms = random_bath('13C', 100, number=2000, seed=10)
        >>> calc = Simulator(1, bath=atoms, r_bath=40, r_dipole=6,
        >>>                  order=2, D=2.88 * 1e6,
        >>>                  magnetic_field=500, pulses=1)
        >>> print(calc)
        Simulator for center array of size 1.
        Parameters of cluster expansion:
        r_bath: 40
        r_dipole: 6
        order: 2

        Bath consists of 549 spins.

        Clusters include:
        549  clusters of order 1.
        457  clusters of order 2.


    Args:

        spin (CenterArray or float or array with shape (n,)):
            CenterArray containing properties of all central spins.

            *OR*

            Total spin of the central spin (Assumes one central spin).

            *OR*

            Array of total spins of the central spins (Assumes *n* central spins).

        position (ndarray):
            Cartesian coordinates ar array of coordinates in Angstrom of the central spin(s).
            Default (0., 0., 0.). If provided, overrides the position in CenterArray.

        alpha (float or ndarray with shape (S, )): :math:`\ket{0}` state of the qubit in :math:`S_z`
            basis or the index of eigenstate to be used as one.

            Default: Lowest energy eigenstate of the central spin Hamiltonian.

            If provided, overrides the alpha state in the CenterArray.

        beta (float or ndarray with shape (S, )): :math:`\ket{1}` state of the qubit in :math:`S_z` basis
            or the index of the eigenstate to be used as one.

            Default: Second lowest energy eigenstate of the central spin Hamiltonian.

            If provided, overrides the beta state in the CenterArray.

        gyro (float or ndarray with shape (3, 3)): Gyromagnetic ratio of the central spin(s) in rad / ms / G.

            *OR*

            Tensor describing central spin interactions with the magnetic field.

            Default -17608.597050 kHz * rad / G - gyromagnetic ratio of the free electron spin.

            If provided, overrides the gyro value in CenterArray.


        D (float or ndarray with shape (3, 3)): D (longitudinal splitting) parameter of central spin
            in ZFS tensor of central spin in kHz.

            *OR*

            Total ZFS tensor. Default 0.

            If provided, overrides the ZFS value in CenterArray.

        E (float): E (transverse splitting) parameter of central spin in ZFS tensor of central spin in kHz.
            Default 0. Ignored if ``D`` is None or tensor.

        bath (ndarray or str): First positional argument of the ``Simulator.read_bath`` method.

            Either:

            - Instance of BathArray class;
            - ndarray with ``dtype([('N', np.unicode_, 16), ('xyz', np.float64, (3,))])`` containing names
              of bath spins (same ones as stored in self.ntype) and positions of the spins in angstroms;
            - the name of the .xyz text file containing 4 columns: name of the bath spin and xyz coordinates in A.

        r_dipole (float): Maximum connectivity distance between two bath spins.

        order (int): Maximum size of the cluster to be considered in CCE expansion.

        n_clusters (dict): Dictionary which contain maximum number of clusters of the given size.
            Has the form ``n_clusters = {order: number}``, where ``order`` is the size of the cluster,
            ``number`` is the maximum number of clusters with this size.

            If provided, sort the clusters by the "strength" of cluster.
            Then the strongest ``number`` of clusters is taken.

            We define the strength of the cluster :math:`s` as an inverse of the sum
            over inverse pairwise interaction strengths of the minimal cluster:

            .. math::

                s = (\sum_{i<j\in C} \frac{r^3}{\gamma_i\gamma_j})^{-1}

            Where :math:`\gamma_i` is the gyromagnetic ration of a spin :math:`i`, :math:`r` is the distance between
            two spins, and the summation of :math:`i, j` goes only over the edges of the minimally connected cluster.

            We define minimally connected cluster as a cluster with lowest possible number of edges that still forms
            a connected graph. If multiple choices of the minimally connected cluster
            for the same cluster are possible, the one with the larger strength :math:`s` is chosen.


        pulses (list or int or Sequence): Number of pulses in CPMG sequence or list with pulses.

        **bath_kw: Additional keyword arguments for the ``Simulator.read_bath`` method.

    """

    def __init__(self, spin, position=None, alpha=None, beta=None, gyro=None, magnetic_field=None,
                 D=None, E=0., r_dipole=None, order=None, bath=None, pulses=None, as_delay=False, n_clusters=None,
                 **bath_kw):

        self.center = None
        """CenterArray: Array of central spins."""
        if isinstance(spin, CenterArray):
            self.center = spin

            if position is not None:
                self.center.xyz = position
            if gyro is not None:
                self.center.gyro = gyro
            if D is not None:
                self.center.set_zfs(D, E)
            if alpha is not None:
                self.alpha = alpha
            if beta is not None:
                self.beta = beta
        else:

            if position is None:
                position = np.zeros(3)
            if gyro is None:
                gyro = ELECTRON_GYRO
            if D is None:
                D = 0.

            if alpha is None:
                alpha = 0
            if beta is None:
                beta = 1

            self.center = CenterArray(spin=spin, gyro=gyro, position=position, D=D, E=E,
                                      alpha=alpha, beta=beta)
        self._magnetic_field = None
        self.magnetic_field = magnetic_field

        self._r_dipole = r_dipole
        self._n_clusters = n_clusters
        self._order = order
        self.clusters = None
        """dict: Dictionary containing information about cluster structure of the bath.
        
        Each keys n correspond to the size of the cluster.
        Each ``Simulator.clusters[n]`` contains ``ndarray`` of shape (m, n),
        where m is the number of clusters of given size, n is the size of the cluster.
        Each row  of this array contains indexes of the bath spins included in the given cluster.
        Generated during ``.generate_clusters`` call."""
        # Bath setting up

        # Properties of the total bath
        self.total_bath = None
        # self.total_imap = None
        # The reduced bath
        self._r_bath = None
        self._bath = None
        # self.imap = None
        self._hyperfine = None
        # External bath properties
        self._external_bath = None
        self._ext_r_bath = None
        self._error_range = 0.2

        self.read_bath(bath=bath, **bath_kw)

        # Parameters of the calculations
        self.pulses = pulses

        self.as_delay = as_delay
        """bool: True if time points are delay between pulses (for equispaced pulses),
        False if time points are total time. Ignored if ``pulses`` contains the time delays."""
        # Initial entangled state of the qubit

        # hybrid CCE
        self.interlaced = False
        """bool: True if use hybrid CCE approach - for each cluster sample over states of the supercluster."""
        # Parameters of MC states
        self.seed = None
        """int: Seed for random number generator, used in random bath states sampling."""
        self.nbstates = None
        """int: Number or random bath states to sample over."""
        self.fixstates = None
        r"""dict: If not None, shows which bath states to fix in random bath states.
        
        Each key is the index of bath spin,
        value - fixed :math:`\hat S_z` projection of the mixed state of nuclear spin."""
        self.masked = None
        """bool: True if mask numerically unstable points (with coherence > 1)
        in the averaging over bath states. 

        .. note::

            It is up to user to check whether the possible instability is due to numerical error
            or unphysical assumptions of the calculations."""
        # Parameters of conventional CCE
        self.second_order = None
        """bool: True if add second order perturbation theory correction to the cluster Hamiltonian in conventional CCE.
        Relevant only for conventional CCE calculations."""
        self.level_confidence = None
        """float: Maximum fidelity of the qubit state to be considered eigenstate of the
        central spin Hamiltonian when ``second_order`` set to True."""
        self.projected_bath_state = None
        """ndarray with shape (n,): Array with z-projections of the bath spins states."""
        self.bath_state = None
        """bath_state (ndarray): Array of bath states."""
        self.timespace = None
        """timespace (ndarray with shape (n,)): Time points at which compute the desired property."""
        self.runner = None
        self.fulldm = False
        self.normalized = True

        self.i = None
        self.j = None

    def __repr__(self):
        bm = (f"Simulator for center array of size {len(self.center)}.\n"
              f"magnetic field:\n{self.magnetic_field.__repr__()}\n\n"
              f"Parameters of cluster expansion:\n"
              f"r_bath: {self.r_bath}\n"
              f"r_dipole: {self.r_dipole}\n"
              f"order: {self.order}\n"
              )

        if self.bath is not None:
            bm += f"\nBath consists of {self.bath.size} spins.\n"

        if self.clusters is not None:
            m = "\nClusters include:\n"
            for k in sorted(self.clusters.keys()):
                m += f"{self.clusters[k].shape[0]}  clusters of order {k}.\n"
            bm += m

        return bm

    @property
    def alpha(self):
        r"""
        ndarray or int: Returns .center.alpha property: :math:`\ket{0}` qubit state of the central spin in Sz basis
        **OR** index of the energy state to be considered as one.
        """
        return self.center.alpha

    @alpha.setter
    def alpha(self, alpha_state):
        self.center.alpha = alpha_state

    @property
    def beta(self):
        r"""
        ndarray or int: Returns .center.beta property: :math:`\ket{1}` qubit state of the central spin in Sz basis
        **OR** index of the energy state to be considered as one.
        """
        return self.center.beta

    @beta.setter
    def beta(self, beta_state):
        self.center.beta = beta_state

    @property
    def magnetic_field(self):
        """
        ndarray: Array containing external magnetic field as (Bx, By, Bz)
        or callable with signature ``magnetic_field(pos)``, where ``pos`` is an array with shape (3,) with the
        position of either bath or central spin. Default is (0, 0, 0).
        """
        return self._magnetic_field

    @magnetic_field.setter
    def magnetic_field(self, magnetic_field):
        self.set_magnetic_field(magnetic_field)

    @property
    def order(self):
        """
        int: Maximum size of the cluster.
        """
        return self._order

    @order.setter
    def order(self, order):
        self._order = order
        self.generate_clusters(order=order)

    @property
    def n_clusters(self):
        """
        dict: Dictionary which contain maximum number of clusters of the given size.
        Has the form ``n_clusters = {order: number}``, where ``order`` is the size of the cluster,
        ``number`` is the maximum number of clusters with this size.

        If provided, sorts the clusters by the strength of cluster interaction,
        equal to the inverse of a sum of inverse pairwise interaction in the minimally connected cluster.
        Then the strongest ``number`` of clusters is taken.
        """
        return self._n_clusters

    @n_clusters.setter
    def n_clusters(self, n_clusters):

        self._n_clusters = n_clusters
        self.generate_clusters(n_clusters=n_clusters)

    @property
    def r_dipole(self):
        """
        float: Maximum connectivity distance.
        """
        return self._r_dipole

    @r_dipole.setter
    def r_dipole(self, r_dipole):
        self._r_dipole = r_dipole
        self.generate_clusters(r_dipole=r_dipole)

    @property
    def pulses(self):
        """
        Sequence: List-like object, containing the sequence of the instantaneous ideal control pulses.

        Each item is ``Pulse`` object, containing the following attributes:

        * **which** (*array-like*): Indexes of the central spins to be rotated by the pulse. Default is all.

        * **x** (*float*): Rotation angle of the central spin about x-axis in radians.

        * **y** (*float*): Rotation angle of the central spin about y-axis in radians.

        * **z** (*float*): Rotation angle of the central spin about z-axis in radians.

        * **delay** (*float or ndarray*): Delay before the pulse or array of delays
          with the same shape as time points.

        Additionally, act as a container object for the pulses, applied to the bath.

        The bath pulses can be accessed as items of the ``Pulse`` object, with name of the item corresponding
        to the name of the bath spin impacted, and the item corresponding to the ``BasePulse`` object with attributes:

            * **x** (*float*): Rotation angle of the central spin about x-axis in radians.

            * **y** (*float*): Rotation angle of the central spin about y-axis in radians.

            * **z** (*float*): Rotation angle of the central spin about z-axis in radians.

        Examples:

            >>> p = Pulse('x', 'pi')
            >>> print(p)
            Pulse((x: 3.14, y: 0.00, z: 0.00))
            >>> pb = Pulse('x', 'pi', bath_names=['13C', '14C'])
            >>> print(pb)
            Pulse((x: 3.14, y: 0.00, z: 0.00), {13C: (x: 3.14, y: 0.00, z: 0.00),
                                                14C: (x: 3.14, y: 0.00, z: 0.00)})
            >>> print(pb['13C'])
            (x: 3.14, y: 0.00, z: 0.00)

        If delay is not provided in **all** pulses, assumes even delay of CPMG sequence.
        If only **some** delays are provided, assumes 0 delay in the pulses without delay provided.

        For the full list of properties, see ``Pulse`` and ``Sequence`` documentations.
        """
        return self._pulses

    @pulses.setter
    def pulses(self, pulses):

        try:
            pulses = ([('x', np.pi), ('y', np.pi)] * ((pulses + 1) // 2))[:pulses]
        except TypeError:
            pulses = pulses

        self._pulses = Sequence(pulses)

    # Properties of the bath
    @property
    def r_bath(self):
        """
        float or array-like: Cutoff size of the spin bath. If ``len(r_bath) > 1``, uses different cutoff
        sizes for each of the central spins. The total bath then is the sum of all bath spins, that are
        close to at least one of the central spins.
        """
        return self._r_bath

    @r_bath.setter
    def r_bath(self, r_bath):
        self.read_bath(r_bath=r_bath)

    @property
    def external_bath(self):
        """
        BathArray: Array with spins read from DFT output (see ``pycce.io``).
        """
        return self._external_bath

    @external_bath.setter
    def external_bath(self, external_bath):
        self.read_bath(external_bath=external_bath)

    @property
    def ext_r_bath(self):
        """
        float: Maximum distance from the central spins of the bath spins
            for which to use the data from ``external_bath``.
        """
        return self._ext_r_bath

    @ext_r_bath.setter
    def ext_r_bath(self, ext_r_bath):
        self.read_bath(ext_r_bath=ext_r_bath)

    @property
    def error_range(self):
        """
        float: Maximum distance between positions in bath and external
            bath to consider two positions the same (default 0.2).
        """
        return self._error_range

    @error_range.setter
    def error_range(self, error_range):
        self.read_bath(error_range=error_range)

    @property
    def hyperfine(self):
        """
        str, func, or Cube instance: This argument tells the code how to generate hyperfine couplings.
            If (``hyperfine = None`` and all A in provided bath are 0) or (``hyperfine = 'pd'``),
            use point dipole approximation. Otherwise can be an instance of ``Cube`` object,
            or callable with signature:

                ``func(array)``

            where array is the ``BathArray`` object.
        """
        return self._hyperfine

    @hyperfine.setter
    def hyperfine(self, hyperfine):
        self.read_bath(hyperfine=hyperfine)

    @property
    def bath(self):
        """
        BathArray: Array of bath spins used in CCE simulations.
        """
        return self._bath

    @bath.setter
    def bath(self, bath_array):
        try:
            self._bath = bath_array.view(BathArray)
        except AttributeError as e:
            print('Bath array should be ndarray or a subclass')
            raise e

        self.total_bath = self._bath

        self._r_bath = None
        self._external_bath = None
        self._ext_r_bath = None
        self.hyperfine = None

    def set_zfs(self, D=None, E=0):
        """
        Set Zero Field Splitting of the central spin from longitudinal ZFS *D* and transverse ZFS *E*.

        Args:
            D (float or ndarray with shape (3, 3)): D (longitudinal splitting) parameter of central spin
                in ZFS tensor of central spin in kHz.

                **OR**

                Total ZFS tensor. Default 0.

            E (float): E (transverse splitting) parameter of central spin in ZFS tensor of central spin in kHz.
                 Default 0. Ignored if ``D`` is None or tensor.
        """

        self.center.set_zfs(D, E)

    def set_magnetic_field(self, magnetic_field=None):
        """
        Set magnetic field from either value of the magnetic field along z-direction or full magnetic field vector.

        Args:
            magnetic_field (float or array-like): Magnetic field along z-axis.

                **OR**

                Array containing external magnetic field as (Bx, By, Bz). Default (0, 0, 0).

        """
        if magnetic_field is None:
            magnetic_field = 0

        if callable(magnetic_field):
            self._magnetic_field = magnetic_field
        else:
            magnetic_field = np.asarray(magnetic_field, dtype=np.float64)

            if magnetic_field.size == 1:
                magnetic_field = np.array([0, 0, magnetic_field.flatten()[0]])

            assert magnetic_field.size == 3, "Improper magnetic field format."

            self._magnetic_field = magnetic_field

    def generate_clusters(self, order=None, r_dipole=None, r_inner=0, strong=False, ignore=None, n_clusters=None):
        r"""
        Generate set of clusters used in CCE calculations.

        The clusters are generated from the following procedure:

        * Each bath spin :math:`i` forms a cluster of one.

        * Bath spins :math:`i` and :math:`j` form cluster of two if there is an edge between them
          (distance :math:`d_{ij} \le` ``r_dipole``).

        * Bath spins :math:`i`, :math:`j`, and :math:`j` form a cluster of three if enough edges connect them
          (e.g., there are two edges :math:`ij` and :math:`jk`)

        * And so on.

        In general, we assume that spins :math:`\{i..n\}` form clusters if they form a connected graph.
        Only clusters up to the size imposed by the ``order`` parameter (equal to CCE order) are included.

        Args:
            order (int): Maximum size of the cluster.

            r_dipole (float): Maximum connectivity distance.

            r_inner (float): Minimum connectivity distance.

            strong (bool):
                True -  generate only clusters with "strong" connectivity (all nodes should be interconnected).
                Default False.

            ignore (list or str, optional): If not None, includes the names of bath spins
                which are ignored in the cluster generation.

            n_clusters (dict): Dictionary which contain maximum number of clusters of the given size.
                Has the form ``n_clusters = {order: number}``, where ``order`` is the size of the cluster,
                ``number`` is the maximum number of clusters with this size.

                If provided, sort the clusters by the strength of cluster interaction,
                Then the strongest ``number`` of clusters is taken.

                Strength of the cluster :math:`s` is defined as
                an inverse of a sum of inverse pairwise interactions of the minimal cluster:

                .. math::

                    s = (\sum_{i<j\in C} \frac{r^3}{\gamma_i\gamma_j})^{-1}

        Returns:

            dict: View of ``Simulator.clusters``. ``Simulator.clusters`` is a dictionary
                with keys corresponding to size of the cluster.

                I.e. ``Simulator.clusters[n]`` contains ``ndarray`` of shape (m, n),
                where m is the number of clusters of given size,
                n is the size of the cluster.
                Each row contains indexes of the bath spins included in the given cluster."""

        self._order = order if order is not None else self._order
        self._r_dipole = r_dipole if r_dipole is not None else self._r_dipole
        self._n_clusters = n_clusters if n_clusters is not None else self._n_clusters

        self.clusters = None

        if self.r_dipole is None:
            raise ValueError('r_dipole was not set')
        elif self.order is None:
            raise ValueError('order was not set')

        clusters = generate_clusters(self.bath, self.r_dipole, self.order,
                                     r_inner=r_inner, strong=strong, ignore=ignore, nclusters=self.n_clusters)

        self.clusters = clusters

        return self.clusters

    def read_bath(self, bath=None, r_bath=None,
                  skiprows=1,
                  external_bath=None,
                  hyperfine=None,
                  types=None,
                  error_range=None,
                  ext_r_bath=None,
                  imap=None,
                  func_kw=None):
        r"""
        Read spin bath from the file or from the ``BathArray``.

        Args:
            bath (ndarray, BathArray or str): Either:

                * Instance of BathArray class;
                * ndarray with ``dtype([('N', np.unicode_, 16), ('xyz', np.float64, (3,))])`` containing names
                  of bath spins (same ones as stored in self.ntype) and positions of the spins in angstroms;
                * the name of the xyz text file containing 4 cols: name of the bath spin and xyz coordinates in A.

            r_bath (float or array-like): Cutoff size of the spin bath. If ``len(r_bath) > 1``, uses different cutoff
                sizes for each of the central spins. The total bath then is the sum of all bath spins, that are
                close to at least one of the central spins.

            skiprows (int, optional): If ``bath`` is name of the file, this argument
                gives number of rows to skip while reading the .xyz file (default 1).

            external_bath (BathArray, optional):
                BathArray containing spins read from DFT output (see ``pycce.io``).

            hyperfine (str, func, or Cube instance, optional):
                This argument tells the code how to generate hyperfine couplings.

                If (``hyperfine = None`` and all A in provided bath are 0) or (``hyperfine = 'pd'``),
                use point dipole approximation.

                Otherwise can be an instance of ``Cube`` object,
                or callable with signature::

                    func(array, *args, **kwargs)

                where ``array`` is array of the bath spins,

            func_kw (dict): Additional keywords if for generating hyperfine couplings if ``hyperfine`` is callable.

            types (SpinDict): SpinDict or input to create one.
                Contains either SpinTypes of the bath spins or tuples which will initialize those.

                See ``pycce.bath.SpinDict`` documentation for details.

            error_range (float, optional): Maximum distance between positions in bath and external
                bath to consider two positions the same (default 0.2).

            ext_r_bath (float, optional): Maximum distance from the central spins of the bath spins
                for which to use the DFT positions.

            imap (InteractionMap): Instance of ``InteractionMap`` class, containing interaction tensors for bath spins.
                Each key of the ``InteractionMap`` is a tuple with indexes of two bath spins.
                The value is the 3x3 tensor describing the interaction between two spins in a format:

                .. math::

                    I^iJI^j = I^i_{x}J_{xx}I^j_{x} + I^i_{x}J_{xy}I^j_{y} ...

        .. note::

            For each bath spin pair without interaction tensor in ``imap``, coupling is approximated assuming
            magnetic point dipole–dipole interaction.  If ``imap = None`` all interactions between bath spins
            are approximated in this way. Then interaction tensor between spins `i` and `j` is computed as:

            .. math::

                \mathbf{J}_{ij} = -\gamma_{i} \gamma_{j} \frac{\hbar^2}{4\pi \mu_0}
                                   \left[ \frac{3 \vec{r}_{ij} \otimes \vec{r}_{ij} - |r_{ij}|^2 I}{|r_{ij}|^5} \right]

            Where :math:`\gamma_{i}` is gyromagnetic ratio of `i` spin, :math:`I` is 3x3 identity matrix, and
            :math:`\vec{r}_{ij}` is distance between two spins.

        Returns:
            BathArray: The view of ``Simulator.bath`` attribute, generated by the method.
        """

        if func_kw is None:
            func_kw = {}

        self._bath = None

        if bath is not None:
            self.total_bath = read_xyz(bath, skiprows=skiprows, spin_types=types, imap=imap)

        bath = self.total_bath

        if bath is None:
            return

        self._r_bath = r_bath if r_bath is not None else self._r_bath
        self._external_bath = external_bath if external_bath is not None else self._external_bath
        self._ext_r_bath = ext_r_bath if ext_r_bath is not None else self._ext_r_bath
        self._hyperfine = hyperfine if hyperfine is not None else self._hyperfine
        self._error_range = error_range if error_range is not None else self._error_range

        bath = bath.expand(self.center.size)

        if self.r_bath is not None:
            mask = False
            rbs = np.broadcast_to(self.r_bath, self.center.size)
            for rb, c in zip(rbs, self.center):
                mask += np.linalg.norm(bath.xyz - np.asarray(c.xyz), axis=-1) < rb
            bath = bath[mask]

        if self.ext_r_bath is not None and self.external_bath is not None:

            where = False
            for c in self.center:
                where += np.linalg.norm(self.external_bath.xyz - c.xyz, axis=1) <= self.ext_r_bath

            self._external_bath = self.external_bath[where]

        if self._hyperfine == 'pd' or (self._hyperfine is None and not bath.A.any()):
            bath = bath.from_center(self.center)

            self._hyperfine = 'pd'

        elif isinstance(self._hyperfine, Cube):
            bath = bath.from_center(self.center, cube=self._hyperfine)

        elif self._hyperfine:
            bath = bath.from_func(self._hyperfine, **func_kw)

        if self.external_bath is not None:
            bath = bath.update(self.external_bath, error_range=self._error_range, ignore_isotopes=True,
                               inplace=True)

        if not bath.size:
            warnings.warn('Provided bath is empty.', stacklevel=2)

        self._bath = bath

        if self.r_dipole and self.order:
            assert self.bath is not None, "Bath spins were not provided to compute clusters"
            self.generate_clusters()

        return self.bath

    @_add_args(_args + _returns)
    def compute(self, timespace, quantity='coherence', method='cce', **kwargs):
        r"""
        General function for computing properties with CCE.

        The dynamics are simulated using the Hamiltonian:

        .. math::

            &\hat H_S = \mathbf{SDS} + \mathbf{B\gamma}_{S}\mathbf{S} \\
            &\hat H_{SB} = \sum_i \mathbf{S}\mathbf{A}_i\mathbf{I}_i \\
            &\hat H_{B} = \sum_i{\mathbf{I}_i\mathbf{P}_i \mathbf{I}_i +
                           \mathbf{B}\mathbf{\gamma}_i\mathbf{I}_i} +
                           \sum_{i>j} \mathbf{I}_i\mathbf{J}_{ij}\mathbf{I}_j

        Here :math:`\hat H_S` is the central spin Hamiltonian with Zero Field splitting tensor :math:`\mathbf{D}` and
        gyromagnetic ratio tensor :math:`\mathbf{\gamma_S} = \mu_S \mathbf{g}`
        are read from ``Simulator.zfs`` and ``Simulator.gyro`` respectively.

        The :math:`\hat H_{SB}` is the Hamiltonian describing interactions between central spin and the bath.
        The hyperfine coupling tensors :math:`\mathbf{A}_i` are read from the ``BathArray`` stored in
        ``Simulator.bath['A']``. They can be generated using point dipole approximation or provided
        by the user (see ``Simulator.read_bath`` for details).

        The :math:`\hat H_{B}` is the Hamiltonian describing interactions between the bath spins.
        The self interaction tensors :math:`\mathbf{P}_i` are read from the ``BathArray`` stored in
        ``Simulator.bath['Q']`` and have to be provided by the user.

        The gyromagnetic ratios :math:`\mathbf{\gamma}_i` are read from the ``BathArray.gyros`` attribuite,
        which is generated from the properties of the types of bath spins, stored in ``BathArray.types``. They can
        either be provided by user or read from the ``pycce.common_isotopes`` object.

        The interaction tensors :math:`\mathbf{J}_{ij}` are assumed from point dipole approximation
        or can be provided  in ``BathArray.imap`` attrubite.

        .. note ::

            The ``compute`` method takes two keyword arguments to determine which quantity to compute and how:

                * `method` can take 'cce' or 'gcce' values, and determines
                  which method to use - conventional or generalized CCE.
                * `quantity` can take 'coherence' or 'noise' values, and determines
                  which quantity to compute - coherence function
                  or autocorrelation function of the noise.

            Each of the methods can be performed with monte carlo bath state sampling
            (if ``nbstates`` keyword is non zero)
            and with interlaced averaging (If ``interlaced`` keyword is set to ``True``).

        Examples:

            First set up Simulator object using random bath of 1000 13C nuclear spins.

                >>> import pycce as pc
                >>> import numpy as np
                >>> atoms = pc.random_bath('13C', 100, number=2000, seed=10) # Random spin bath
                >>> calc = pc.Simulator(1, bath=atoms, r_bath=40, r_dipole=6,
                >>>                     order=2, D=2.88 * 1e6, # D of NV in GHz -> kHz
                >>>                     magnetic_field=500, pulses=1)
                >>> ts = np.linspace(0, 2, 101) # timesteps

            We set magnetic field to 500 G along z-axis and chose 1 decoupling pulse (Hahn-echo) in this example.
            The zero field splitting is set to the one of NV center in diamond.

            Run conventional CCE calculation at time points ``timespace`` to obtain coherence without second order
            effects:

                >>> calc.compute(ts)

            This will call ``Simulator.cce_coherence`` method with default keyword values.

            Compute the coherence conventional CCE coherence with second order interactions between bath spins:

                >>> calc.compute(ts, second_order=True)

            Compute the coherence with conventional CCE with bath state sampling (over 10 states):

                >>> calc.compute(ts, nbstates=10)

            Compute the coherence with generalized CCE:

                >>> calc.compute(ts, method='gcce')

            This will call ``Simulator.gcce_dm`` method with default keyword values and obtain off diagonal
            element as :math:`\bra{0} \hat \rho_{C} \ket{1}`,
            where :math:`\hat \rho_{C}` is the density matrix of the qubit.

            Compute the coherence with generalized CCE with bath state sampling (over 10 states):

                >>> calc.compute(ts, method='gcce', nbstates=10)

        Args:

            timespace (ndarray with shape (n,)): Time points at which compute the desired property.

            quantity (str): Which quantity to compute. Case insensitive.

                Possible values:

                    - 'coherence': compute coherence function.
                    - 'noise': compute noise autocorrelation function.

            method (str): Which implementation of CCE to use. Case insensitive.

                Possible values:

                    - 'cce': conventional CCE, where interactions are mapped on 2 level pseudospin.
                    - 'gcce': Generalized CCE where central spin is included in each cluster.
        """
        if quantity.lower() == 'noise':
            kwargs.setdefault('masked', None)

        else:
            kwargs.setdefault('masked', True)

        self.timespace = timespace
        self._prepare(**kwargs)

        runner = self._compute_func[method.lower()][quantity.lower()].from_simulator(self)
        self.runner = runner

        if not self.interlaced:

            if self.nbstates:
                result = runner.sampling_run()

            else:
                result = runner.run()

        else:

            if self.nbstates:
                result = runner.sampling_interlaced_run()

            else:
                result = runner.interlaced_run()

        return result

    def _broadcast(self):
        """
        Update attributes for the ``Simulator`` object from the root process.
        """
        calc = _broadcast_simulator(self)
        dict_with_attr = vars(calc)

        for key in dict_with_attr:
            setattr(self, key, dict_with_attr[key])

    _compute_func = {
        'cce': {
            'coherence': CCE,
            'noise': CCENoise
        },
        'gcce': {
            'coherence': gCCE,
            'noise': gCCENoise,
        }
    }

    @_add_args(_args)
    def _prepare(self,
                 pulses=None,
                 i=None,
                 j=None,
                 normalized=True,
                 magnetic_field=None,
                 as_delay=False,
                 nbstates=None,
                 seed=None,
                 masked=True,
                 parallel_states=False,
                 fixstates=None,
                 second_order=False,
                 level_confidence=0.95,
                 direct=False,
                 parallel=False,
                 interlaced=False,
                 fulldm=False):
        """

        Args:

        """
        self.i = i
        self.j = j

        if pulses is not None:
            self.pulses = pulses
        if magnetic_field is not None:
            self.set_magnetic_field(magnetic_field)

        self.parallel = parallel
        self.parallel_states = parallel_states
        self.as_delay = as_delay
        self.direct = direct
        self.normalized = normalized

        self.second_order = second_order
        self.level_confidence = level_confidence
        self.fixstates = fixstates

        self.nbstates = nbstates
        self.seed = seed
        self.masked = masked
        self.interlaced = interlaced
        self.fulldm = fulldm

        if parallel or parallel_states:
            self._broadcast()

        # if bath_state is not None:
        #     self.projected_bath_state = project_bath_states(bath_state)

        # else:
        #     self.projected_bath_state = None


def _broadcast_simulator(simulator=None, root=0):
    """
    Broadcast ``Simulator`` object from root to all mpi processes.

    Args:

        simulator (Simulator): Object to broadcast. Should be defined at root process.
        root (int): Index of the root process.

    Returns:

        Simulator: Broadcasted object.
    """
    try:
        import mpi4py
    except ImportError:
        raise RuntimeError('Could not find mpi4py. Cannot broadcast the Simulator')

    comm = mpi4py.MPI.COMM_WORLD

    # size = comm.Get_size()
    rank = comm.Get_rank()

    if rank == root:
        bath = simulator.bath
        center = simulator.center

        simulator._bath = None
        simulator.center = None

    else:
        bath = None
        center = None
        simulator = None

    nbath = broadcast_array(bath, root=root)
    ncenter = broadcast_array(center, root=root)

    nsim = comm.bcast(simulator, root=root)

    nsim.center = ncenter
    nsim._bath = nbath
    return nsim



import copy
import warnings
from collections import UserDict
from collections.abc import Mapping

import numpy as np
from numpy.lib.recfunctions import repack_fields
from pycce.bath.map import InteractionMap
from pycce.bath.state import BathState
from pycce.constants import HBAR_MU0_O4PI, ELECTRON_GYRO, HBAR_SI, NUCLEAR_MAGNETON, PI2
from pycce.utilities import gen_state_list, vector_from_s, rotate_coordinates, rotate_tensor, _add_args

HANDLED_FUNCTIONS = {}

_set_str_kinds = {'U', 'S'}

_stevens_str_doc = r"""
            dict: Dictionary with additional spin Hamiltonian parameters.
            Key denotes the product of spin operators as:
            
            Either a string containing ``x, y, z, +, -``  where each symbol is a corresponding spin operator:
            
                - ``x`` == :math:`S_x`
                - ``y`` == :math:`S_y`
                - ``z`` == :math:`S_z`
                - ``p`` == :math:`S_+`
                - ``m`` == :math:`S_-`
            
            Several symbols is a product of those spin operators.
            
            Or a tuple with indexes (k, q) for Stevens operators
            (see https://www.easyspin.org/documentation/stevensoperators.html).
            
            The item is the coupling parameter in float.
            
            Examples:
            
                - ``d['pm'] = 2000`` corresponds to the Hamiltonian term
                  :math:`\hat H_{add} = A \hat S_+ \hat S_-` with :math:`A = 2` MHz.
            
                - ``d[2, 0] = 1.5e6`` corresponds to Stevens operator
                  :math:`B^q_k \hat O^q_k = 3 \hat S_z - s(s+1) \hat I`
                  with :math:`k = 2`, :math:`q = 0`, and :math:`B^q_k = 1.5` GHz. """


class BathArray(np.ndarray):
    """
    Subclass of ``ndarray`` containing information about the bath spins.

    The subclass has fixed structured datatype::

         _dtype_bath = np.dtype([('N', np.unicode_, 16),
                                 ('xyz', np.float64, (3,)),
                                 ('A', np.float64, (3, 3)),
                                 ('Q', np.float64, (3, 3))])

    Accessing different fields results in the ``ndarray`` view.

    Each of the fields can be accessed as the attribute of the ``BathArray`` instance and modified accordingly.
    In addition to the name fields, the information of the bath spin types is stored in the ``types`` attribute.
    All of the items in ``types`` can be accessed as attributes of the BathArray itself.

    Examples:
        Generate empty ``BathArray`` instance.

        >>> ba = BathArray((3,))
        >>> print(ba)
        [('', [0., 0., 0.], [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]], [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]])
         ('', [0., 0., 0.], [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]], [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]])
         ('', [0., 0., 0.], [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]], [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]])]

        Generate BathArray from the set of arrays:

        >>> import numpy as np
        >>> ca = np.random.random((2, 3))
        >>> sn = ['1H', '2H']
        >>> hf = np.random.random((2, 3, 3))
        >>> ba = BathArray(ca=ca, hf=hf, sn=sn)
        >>> print(ba.N, ba.types)
        ['1H' '2H'] SpinDict(1H: (1H, 0.5, 26.7519), 2H: (2H, 1, 4.1066, 0.00286))

    .. warning::
        Due to how structured arrays work, if one uses a boolean array to access an subarray,
        and then access the name field, the initial array *will not change*.

        Example:

            >>> ha = BathArray((10,), sn='1H')
            >>> print(ha.N)
            ['1H' '1H' '1H' '1H' '1H' '1H' '1H' '1H' '1H' '1H']
            >>> bool_mask = np.arange(10) % 2 == 0
            >>> ha[bool_mask]['N'] = 'e'
            >>> print(ha.N)
            ['1H' '1H' '1H' '1H' '1H' '1H' '1H' '1H' '1H' '1H']

            To achieve the desired result, one should first access the name field and only then apply the boolean mask:

            >>> ha['N'][bool_mask] = 'e'
            >>> print(ha.N)
            ['e' '1H' 'e' '1H' 'e' '1H' 'e' '1H' 'e' '1H']

    Each bath spin can initiallized in some specific state accessing the ``.state`` attribute. It takes
    both state vectors and density matrices as values. See ``.state`` attribute documentation for details.

    Args:
        shape (tuple): Shape of the array.

        array (array-like):
            Either an unstructured array with shape (n, 3) containing coordinates of bath spins as rows OR
            structured ndarray with the same fields as the datatype of the bath.

        name (array-like):
            Array of the bath spin name.

        hyperfines (array-like):
            Array of the hyperfine tensors with shape (n, 3, 3).

        quadrupoles (array-like):
            Array of the quadrupole tensors with shape (n, 3, 3).

        efg (array-like):
            Array of the electric field gradients with shape (n, 3, 3) for each bath spin.
            Used to compute Quadrupole tensors for spins >= 1.
            Requires the spin types either be found in ``common_isotopes`` or specified with ``types`` argument.

        types (SpinDict):
            SpinDict or input to create one.
            Contains either SpinTypes of the bath spins or tuples which will initialize those.
            See ``pycce.bath.SpinDict`` documentation for details.

        imap (InteractionMap):
            Instance of InteractionMap containing user defined
            interaction tensors between bath spins stored in the array.

        ca (array-like):
            Shorthand notation for ``array`` argument.

        sn (array-like):
            Shorthand notation for ``name`` argument.

        hf (array-like):
            Shorthand notation for ``hyperfines`` argument.

        q (array-like):
            Shorthand notation for ``quadrupoles`` argument.

    """

    _dtype_names = ('N', 'xyz', 'A', 'Q')

    def __new__(subtype, shape=None, array=None,
                names=None, hyperfines=None, quadrupoles=None,
                types=None, imap=None,
                ca=None, sn=None, hf=None, q=None, efg=None, state=None,
                center=1):
        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments. This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to BathArray.__array_finalize__
        if center == 1:
            atupl = (3, 3)
        else:
            atupl = (center, 3, 3)

        _dtype_bath = np.dtype([('N', np.unicode_, 16),
                                ('xyz', np.float64, (3,)),
                                ('A', np.float64, atupl),
                                ('Q', np.float64, (3, 3)),
                                # ('state', object),
                                # ('proj', np.float64),
                                ])

        if array is None and ca is not None:
            array = ca

        if names is None and sn is not None:
            names = sn

        if hyperfines is None and hf is not None:
            hyperfines = hf

        if quadrupoles is None and q is not None:
            quadrupoles = q

        if shape is not None:
            # set the new 'info' attribute to the value passed
            obj = super(BathArray, subtype).__new__(subtype, shape, dtype=_dtype_bath)
        else:

            for a in (array, hyperfines, quadrupoles):
                if a is not None:
                    try:
                        obj = super(BathArray, subtype).__new__(subtype, (np.asarray(a).shape[0],),
                                                                dtype=_dtype_bath)
                    except IndexError:  # Empty tuple
                        obj = super(BathArray, subtype).__new__(subtype, np.asarray(a).shape,
                                                                dtype=_dtype_bath)
                    break
            else:
                raise ValueError('No shape provided')

        obj.types = SpinDict()
        obj.imap = imap

        obj._state = BathState(obj.size)

        if state is not None:
            obj.state = state

        # obj.__projected_state = np.zeros(obj.shape, dtype=np.float64)

        if types is not None:
            try:
                obj.add_type(**types)
            except TypeError:
                obj.add_type(*types)

        if array is not None:
            array = np.asarray(array)
            if array.dtype.names is not None:
                for n in array.dtype.names:
                    obj[n] = array[n]
            else:
                obj['xyz'] = array.reshape(-1, 3)

        if names is not None:
            obj['N'] = np.asarray(names).reshape(-1)
        if hyperfines is not None:
            obj['A'] = np.asarray(hyperfines).reshape(-1, *atupl)
        if quadrupoles is not None:
            obj['Q'] = np.asarray(quadrupoles).reshape(-1, 3, 3)
        elif efg is not None:
            obj.from_efg(efg)

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(BathArray, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. InfoArray():
        #    obj is None
        #    (we're in the middle of the BathArray.__new__
        #    constructor, and BathArray stuff will be set when we return to
        #    BathArray.__new__)
        if obj is None: return
        # From view casting - e.g arr.view(BathArray):
        #    obj is arr
        #    (type(obj) can be BathArray)
        # From new-from-template - e.g arr[:3]
        #    type(obj) is BathArray
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'types', because this
        # method sees all creation of default objects - with the
        # BathArray.__new__ constructor, but also with
        # arr.view(BathArray).
        # if obj.dtype.names != self._dtype_names:
        #     warnings.warn('Trying to view array with unknown dtype as BathArray. '
        #                   'This can lead to unexpected results.',
        #                   RuntimeWarning, stacklevel=2)

        self.types = getattr(obj, 'types', None)
        self.imap = getattr(obj, 'imap', None)
        self._state = getattr(obj, '_state', None)
        # We do not need to return anything

    def __array_function__(self, func, types, args, kwargs):

        if func not in HANDLED_FUNCTIONS:
            if not all(issubclass(t, np.ndarray) for t in types):
                # Defer to any non-subclasses that implement __array_function__
                return NotImplemented
            # Use NumPy's private implementation without __array_function__
            # dispatching
            return func._implementation(*args, **kwargs)
        # Note: this allows subclasses that don't override
        # __array_function__ to handle MyArray objects
        if not any(issubclass(t, BathArray) for t in types):
            return NotImplemented

        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    def sort(self, axis=-1, kind=None, order=None):
        """
        Sort array in-place. Is implemented only when imap is None. Otherwise use ``np.sort``.
        """
        if self.imap is None:
            super().sort(axis=axis, kind=kind, order=order)
        else:
            raise NotImplementedError('Inplace sort is implemented only when .imap is None')

    @_add_args(_stevens_str_doc)
    @property
    def h(self):
        return _get_sd_attribute(self, 'h')

    @property
    def name(self):
        """
        ndarray: Array of the ``name`` attribute for each spin in the array from ``types`` dictionary.

        .. note::
            While the value of this attribute should be the same as the ``N`` field of the BathArray instance,
            ``.name`` *should not* be used for production as it creates a *new* array from ``types`` dictionary.
        """
        return _get_sd_attribute(self, 'name')

    @name.setter
    def name(self, initial_value):
        _set_sd_attribute(self, 'name', initial_value)

    @property
    def s(self):
        """
        ndarray: Array of the ``spin`` (spin value) attribute for each spin in the array from ``types`` dictionary.
        """
        return _get_sd_attribute(self, 's')

    @s.setter
    def s(self, initial_value):
        _set_sd_attribute(self, 's', initial_value)

    @property
    def dim(self):
        """
        ndarray: Array of the ``dim`` (dimensions of the spin) attribute
            for each spin in the array from ``types`` dictionary.
        """
        return _get_sd_attribute(self, 'dim')

    @property
    def gyro(self):
        """
        ndarray: Array of the ``gyro`` (gyromagnetic ratio)
            attribute for each spin in the array from ``types`` dictionary.
        """
        return _get_sd_attribute(self, 'gyro')

    @gyro.setter
    def gyro(self, initial_value):
        _set_sd_attribute(self, 'gyro', initial_value)

    @property
    def q(self):
        """
        ndarray: Array of the ``q`` (quadrupole moment)
            attribute for each spin in the array from ``types`` dictionary.
        """
        return _get_sd_attribute(self, 'q')

    @q.setter
    def q(self, initial_value):
        _set_sd_attribute(self, 'q', initial_value)

    @property
    def detuning(self):
        """
        ndarray: Array of the ``detuning``
            attribute for each spin in the array from ``types`` dictionary.
        """
        return _get_sd_attribute(self, 'detuning')

    @detuning.setter
    def detuning(self, initial_value):
        _set_sd_attribute(self, 'detuning', initial_value)

    @property
    def x(self):
        """
        ndarray: Array of x coordinates for each spin in the array (``bath['xyz'][:, 0]``).
        """
        return self['xyz'][..., 0]

    @x.setter
    def x(self, val):
        self['xyz'][..., 0] = val

    @property
    def y(self):
        """
        ndarray: Array of y coordinates for each spin in the array (``bath['xyz'][:, 1]``).
        """
        return self['xyz'][..., 1]

    @y.setter
    def y(self, val):
        self['xyz'][..., 1] = val

    @property
    def z(self):
        """
        ndarray: Array of z coordinates for each spin in the array (``bath['xyz'][:, 2]``).
        """
        return self['xyz'][..., 2]

    @z.setter
    def z(self, val):
        self['xyz'][..., 2] = val

    @property
    def N(self):
        """
        ndarray: Array of name for each spin in the array (``bath['N']``).
        """
        return self['N']

    @N.setter
    def N(self, val):
        self['N'] = val

    @property
    def xyz(self):
        """
        ndarray: Array of coordinates for each spin in the array (``bath['xyz']``).
        """
        return self['xyz']

    @xyz.setter
    def xyz(self, val):
        self['xyz'] = val

    @property
    def A(self):
        """
        ndarray: Array of hyperfine tensors for each spin in the array (``bath['A']``).
        """
        return self['A']

    @A.setter
    def A(self, val):
        self['A'] = val

    @property
    def Q(self):
        """
        ndarray: Array of quadrupole tensors for each spin in the array (``bath['Q']``).
        """
        return self['Q']

    @Q.setter
    def Q(self, val):
        self['Q'] = val

    @property
    def nc(self):
        """int: Number of central spins."""
        selfdim = len(self.shape)
        return self.A.shape[selfdim] if len(self.A.shape) == selfdim + 3 else 1

    @property
    def state(self):
        """
        BathState: Array of the bath spin states.

        Can have three types of entries:

            - **None**. If entry is **None**, assumes fully random density matrix. Default value.
            - **ndarray with shape (s,)**. If entry is vector, corresponds to the pure state of the spin.
            - **ndarray with shape (s, s)**. If entry is a matrix, corresponds to the density matrix of the spin.

        Examples:

            >>> print(ba.state)
            [None None]
            >>> ba[0].state = np.array([0, 1])
            >>> print(ba.state)
            [array([0, 1]) None]
        """
        return self._state

    @state.setter
    def state(self, rho):
        self._state[...] = rho

    @property
    def proj(self):
        """
        ndarray: Array of :math:`S_z` projections of the bath spin states.
        """
        return self.state.proj

    @proj.setter
    def proj(self, rho):
        # rho can be either (on the example of s = 1)
        # projection of pure state:  rho = 1 for ms = 1
        # pure state: rho = [0,0,1] for ms = - 1
        # density matrix rho: = [[0,0,0],[0,1,0],[0,0,0]] for ms = 0
        rho = np.asarray(rho)

        if not rho.shape:
            # assume s is int or float showing the spin projection in the pure state

            d = self.dim
            if not self.shape:
                rho = vector_from_s(rho, d)
            else:
                if (d == d[0]).all():
                    d = d[0]
                    rho = np.broadcast_to(vector_from_s(rho, d), self.shape + (d,))
                else:
                    rho = [vector_from_s(rho, d_i) for d_i in d]
        else:
            rho = gen_state_list(rho, np.broadcast_to(self.dim, self.shape))
        self._state[...] = rho

    @property
    def has_state(self):
        """
        ndarray: Bool array. True if given spin was initialized with a state, False otherwise.
        """
        return self.state.has_state

    def __getitem__(self, item):
        if isinstance(item, (int, np.int32, np.int64)):
            item = (Ellipsis, item)
            obj = np.ndarray.__getitem__(self, item)
            # obj._state = self._state[item]
            obj._state = self._state._get_state(item)

            return obj

        elif isinstance(item, tuple) and not item:
            return np.ndarray.__getitem__(self, item)
        #     else:
        #         item = (Ellipsis,) + item
        #         obj = np.ndarray.__getitem__(self, item)
        #         # obj._state = self._state[item]
        #         return obj

        # if string then return ndarray view of the field
        elif isinstance(item, (str, np.str_)):
            try:
                value = self.view(np.ndarray).__getitem__(item)

                if not value.shape:
                    return value[()]

                return value

            except ValueError:
                return self[self['N'] == item]

        else:

            obj = np.ndarray.__getitem__(self, item)

            try:
                obj._state = self._state._get_state(item)

                if self.imap is not None:
                    if not isinstance(item, tuple):

                        if isinstance(item, slice):
                            item = np.arange(self.size)[item]
                        smap = self.imap.subspace(item)
                        if smap:
                            obj.imap = smap

            except AttributeError:
                pass

            return obj

    def __setitem__(self, key, val):
        np.ndarray.__setitem__(self, key, val)

        if isinstance(val, str):
            if val not in self.types.keys():
                try:
                    self.types[val] = copy.copy(common_isotopes[val])
                except KeyError:
                    warnings.warn(_spin_not_found_message(val), stacklevel=2)
            return

        elif isinstance(val, (int, float, complex)):
            return

        val = np.asarray(val)

        if (val.dtype.names is not None) and ('N' in val.dtype.names):
            val = val['N']

        if val.dtype.kind in _set_str_kinds:
            if not val.shape:
                n = val[()]
                if n not in self.types.keys():
                    try:
                        self.types[n] = copy.copy(common_isotopes[n])
                    except KeyError:
                        warnings.warn(_spin_not_found_message(n), stacklevel=2)
                return

            for n in np.unique(val):
                if n not in self.types.keys():
                    try:
                        self.types[n] = copy.copy(common_isotopes[n])
                    except KeyError:
                        warnings.warn(_spin_not_found_message(n), stacklevel=2)
            return

    def __eq__(self, other):
        try:

            xyzs = (self['xyz'] == other['xyz']).all(axis=-1)
            hfs = (self['A'] == other['A']).reshape(*self.shape, -1).all(axis=-1)
            qds = (self['Q'] == other['Q']).reshape(*self.shape, -1).all(axis=-1)

            return xyzs & hfs & qds

        except (IndexError, TypeError):
            return super().__eq__(other)

    def add_type(self, *args, **kwargs):
        """
        Add spin type to the ``types`` dictionary.

        Args:
            *args: Any number of positional inputs to create ``SpinDict`` entries.
                E.g. the tuples of form (name :obj:`str`, spin :obj:`float`, gyro :obj:`float`, q :obj:`float`).
            **kwargs: Any number of keyword inputs to create ``SpinDict`` entries.
                E.g. name = (spin, gyro, q).

        For details and allowed inputs see ``SpinDict`` documentation.

        Returns:
            SpinDict: A view of ``self.types`` instance.
        """
        self.types.add_type(*args, **kwargs)

    def add_interaction(self, i, j, tensor):
        """
        Add interactions tensor between bath spins with indexes ``i`` and ``j``.

        .. note::

            If called from the subarray this method **does not** change the tensors of the total BathArray.

        Args:
            i (int or ndarray (n,) ):
                Index of the first spin in the pair or array of the indexes of the first spins in n pairs.
            j (int or ndarray with shape (n,)):
                Index of the second spin in the pair or array of the indexes of the second spins in n pairs.
            tensor (ndarray with shape (3,3) or (n, 3,3)):
                Interaction tensor between the spins i and j or array of tensors.

        """
        if self.imap is None:
            self.imap = InteractionMap(i, j, tensor)
            # self.imap[i, j] = tensor
        else:
            self.imap[i, j] = tensor

    def update(self, ext_bath, error_range=0.2, ignore_isotopes=True, inplace=True):
        """
        Update the properties of the spins in the array using data from other ``BathArray`` instance.
        For each spin in ``ext_bath`` check whether there is such spin in the array that has the same position
        within allowed error range given by ``error_range`` and has the same name.
        If such spins is found in the array, then it's coordinates, hyperfine tensor and quadrupole tensor
        are updated using the values of the spin in the ``ext_bath`` object.

        If ``ignore_isotopes`` is true, then the name check ignores numbers in the name of the spins.

        Args:
            ext_bath (BathArray): Array of the new spins.

            error_range (float): +- distance in Angstrom within which two positions are considered to be the same.
                Default is 0.2 A.

            ignore_isotopes (bool): True if ignore numbers in the name of the spins. Default True.

            inplace (bool): True if changes parameters of the array in place. If False, returns copy of the array.

        Returns:
            BathArray: updated BathArray instance.
        """
        bath = update_bath(self, ext_bath, error_range, ignore_isotopes, inplace)
        return bath

    def transform(self, center=None, cell=None, rotation_matrix=None, style='col', inplace=True):
        """
        Coordinate transformation of BathArray.

        Args:
            center (ndarray with shape (3,)): (0, 0, 0) position of new coordinates in the initial frame.

            cell (ndarray with shape (3, 3)): Cell vectors in cartesian coordinates
                if initial coordinates of the ``atoms`` are in crystallographic reference frame.

            rotation_matrix (ndarray with shape (3, 3)): Rotation matrix R of the **coordinate system**.

                E.g. ``R @ [0, 0, 1] = [a, b, c]`` where ``[a, b, c]`` are
                coordinates of the z axis of the new coordinate
                system in the old coordinate system.

                Note, that rotation is applied after transition from cell coordinates to the cartesian coordinates,
                in which cell vectors are stored.

            style (str): Can have two values: 'col' or 'row'.
                Shows how ``cell`` and ``rotation_matrix`` matrices are given:

                    * if 'col', each column of the matrix is a vector in previous coordinates;
                    * if 'row' - each row is a new vector.

                Default is 'col'.

            inplace (bool): If true, makes inplace changes to the provided array.

        Returns:
            BathArray: Transformed array with bath spins.
        """

        bath = transform(self, center=center, cell=cell, rotation_matrix=rotation_matrix,
                         style=style, inplace=inplace)
        return bath

    def from_center(self, center, inplace=True, cube=None, which=0, **kwarg):
        """
        Generate hyperfine couplings using either the point dipole approximation or spin density in the .cube format,
        with the information from the CenterArray instance.

        Args:
            center (CenterArray): Array, containing properties of the central spin

            inplace (bool): True if changes parameters of the array in place. If False, returns copy of the array.

            cube (Cube or iterable of Cubes): An instance of ``Cube`` object,
                which contains spatial distribution of spin density of central spins.
                For details see documentation of ``Cube`` class.

            which (int): If ``cube`` is a single Cube instance,
                this is an index of the central spin it corresponds to.

            **kwarg: Additional arguments for .from_cube method.

        Returns:
            BathArray: Updated BathArray instance.

        """
        if inplace:
            array = self
        else:
            array = self.copy()

        if array.nc != center.size:
            array = array.expand(center.size)
            if inplace:
                warnings.warn("Cannot change array inplace, using a copy instead.")

        if center.size == 1:
            if cube is None:
                array.from_point_dipole(center[0].xyz, center[0].gyro, inplace=True)
            else:
                array.from_cube(cube, center[0].gyro, inplace=True, **kwarg)
        else:
            if cube is None:
                array.from_point_dipole(center.xyz, center.gyro, inplace=True)
            else:
                try:
                    for index, (cen, cub) in enumerate(zip(center, cube)):
                        array.from_cube(cub, cen.gyro, inplace=True, which=index, **kwarg)
                except TypeError:
                    array.from_point_dipole(center.xyz, center.gyro, inplace=True)
                    array.from_cube(cube, center[which].gyro, inplace=True, which=which, **kwarg)

        return array

    def from_point_dipole(self, position, gyro_center=ELECTRON_GYRO, inplace=True):
        """
        Generate hyperfine couplings, assuming that bath spins interaction with central spin is the same as the
        one between two magnetic point dipoles.

        Args:
            position (ndarray with shape (3,)): position of the central spin

            gyro_center (float or ndarray with shape (3,3)):
                gyromagnetic ratio of the central spin

                **OR**

                tensor corresponding to interaction between magnetic field and central spin.

            inplace (bool): True if changes parameters of the array in place. If False, returns copy of the array.

        Returns:
            BathArray: Updated BathArray instance with changed hyperfine couplings.
        """
        if inplace:
            array = self
        else:
            array = self.copy()

        position = np.asarray(position)

        if array.nc > 1:
            position = np.broadcast_to(position, (array.nc, 3))
            gyro_center = np.asarray(gyro_center)
            gyro_center = np.broadcast_to(gyro_center, (array.nc, *gyro_center.shape[1:]))

            for i in range(array.nc):
                pos = array.xyz - position[i]
                array.A[:, i] = point_dipole(pos, array.gyro, gyro_center[i])

            return array

        pos = array.xyz - position
        array.A = point_dipole(pos, array.gyro, gyro_center)

        return array

    def from_cube(self, cube, gyro_center=ELECTRON_GYRO, inplace=True, which=0, **kwargs):
        """
        Generate hyperfine couplings, assuming that bath spins interaction with central spin can be approximated as
        a point dipole, interacting with given spin density distribution.

        Args:
            cube (Cube): An instance of `Cube` object, which contains spatial distribution of spin density.
                For details see documentation of `Cube` class.

            gyro_center (float): Gyromagnetic ratio of the central spin.

            inplace (bool): True if changes parameters of the array in place. If False, returns copy of the array.

        Returns:
            BathArray: Updated BathArray instance with changed hyperfine couplings.
        """
        if inplace:
            array = self
        else:
            array = self.copy()

        gyros = array.gyro

        if array.nc > 1:
            array.A[:, which] = cube.integrate(array.xyz, gyros, gyro_center, **kwargs)

            return array

        array.A = cube.integrate(array.xyz, gyros, gyro_center, **kwargs)

        return array

    def from_func(self, func, *args, inplace=True, **kwargs):
        """
        Generate hyperfine couplings from user-defined function.

        Args:

            func (func):
                Callable with signature::

                    func(array, *args, **kwargs)

                where ``array`` is array of the bath spins,
            *args: Positional arguments of the ``func``.
            **kwargs: Keyword arguments of the ``func``.
            inplace (bool): True if changes parameters of the array in place. If False, returns copy of the array.

        Returns:
            BathArray: Updated BathArray instance with changed hyperfine couplings.

        """
        if inplace:
            array = self
        else:
            array = self.copy()

        func(array, *args, **kwargs)

        return array

    def from_efg(self, efg, inplace=True):
        """
        Generate quadrupole splittings from electric field gradient tensors for spins >= 1.

        Args:
            efg (array-like): Array of the electric field gradients for each bath spin. The data for spins-1/2
                should be included but can be any value.

            inplace (bool): True if changes parameters of the array in place. If False, returns copy of the array.

        Returns:
            BathArray: Updated BathArray instance with changed quadrupole tensors.

        """

        if inplace:
            array = self
        else:
            array = self.copy()

        efg = np.asarray(efg).reshape((-1, 3, 3))

        spins = np.asarray(array.s)
        qmoments = np.asarray(array.q)

        where = spins > 0.5
        if not where.shape:
            if where:
                pref = qmoments / (2 * spins * (2 * spins - 1))
            else:
                pref = 0
        else:
            pref = np.zeros(efg.shape[0], dtype=np.float64)
            pref[where] = qmoments[where] / (2 * spins[where] * (2 * spins[where] - 1))
            pref = pref[..., np.newaxis, np.newaxis]

        array.Q = pref * efg
        return array

    def dist(self, position=None):
        """
        Compute the distance of the bath spins from the given position.

        Args:
            position (ndarray with shape (3,)):
                Cartesian coordinates of the position from which to compute the distance. Default is (0, 0, 0).

        Returns:
            ndarray  with shape (n,): Array of distances of each bath spin from the given position in angstrom.
        """
        if isinstance(position, BathArray):
            return np.linalg.norm(self.xyz - position.xyz, axis=-1)

        if position is None:
            position = np.zeros(3)
        else:
            position = np.asarray(position)

        return np.linalg.norm(self.xyz - position, axis=-1)

    def savetxt(self, filename, fmt='%18.8f', strip_isotopes=False, **kwargs):
        r"""
        Save name of the isotopes and their coordinates to the txt file of xyz format.

        Args:
            filename (str or file): Filename or file handle.
            fmt (str): Format of the coordinate entry.
            strip_isotopes (bool): True if remove numbers from the name of bath spins. Default False.
            **kwargs: Additional keywords of the ``numpy.savetxt`` function.
        """
        kwargs.setdefault('comments', '')
        ar = repack_fields(self.view(np.ndarray)[['N', 'xyz']]).view(
            np.dtype([('N', 'U16'), ('x', 'f8'), ('y', 'f8'), ('z', 'f8')]))

        if strip_isotopes:
            ar['N'] = np.core.defchararray.strip(ar['N'], '1234567890')

        header = f'{ar.size}\n'
        for n, c in zip(*np.unique(ar['N'], return_counts=True)):
            header += f'{n}{c} '

        kwargs.setdefault('header', header)
        np.savetxt(filename, ar, fmt=('%s', fmt, fmt, fmt), **kwargs)

    def expand(self, ncenters):

        array = BathArray(array=self.xyz, quadrupoles=self.Q, names=self.N, center=ncenters, imap=self.imap,
                          types=self.types, state=self.state)

        hyperfine = self.A
        if hyperfine.any():
            ocs = self.nc  # old central spin
            if ocs == ncenters:
                array.A = hyperfine

            elif ocs == 1 and ncenters > 1:
                array.A[..., 0, :, :] = hyperfine

            elif ocs > 1 and ncenters == 1:
                array.A[..., :, :] = hyperfine[..., 0, :, :]

            else:
                limit = ocs if ocs < ncenters else ncenters
                array.A[..., :limit + 1, :, :] = hyperfine[..., :limit + 1, :, :]

        return array


def implements(numpy_function):
    """Register an __array_function__ implementation for BathArray objects."""

    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


@implements(np.sort)
def sort(a, axis=-1, kind=None, order=None):
    """
    Return a sorted copy of an array. Overrides numpy.sort function.
    """
    indexes = np.argsort(a, axis=axis, kind=kind, order=order)

    return a[indexes]


@implements(np.argsort)
def argsort(a, *args, **kwargs):
    """
    Return a indexes of an sorted array. Overrides ``numpy.argsort`` function.
    """
    return np.argsort(a, *args, **kwargs).view(np.ndarray)


@implements(np.delete)
def delete(arr, obj, axis=None):
    newarr = np.delete(arr.view(np.ndarray), obj, axis=axis).view(BathArray)
    newarr._state = BathState(newarr.size)
    newarr.types = arr.types

    if arr.imap:
        newarr.imap = arr.imap.subspace(obj)

    newarr.state = np.delete(arr.state[...], obj, axis=axis)
    return newarr


@implements(np.concatenate)
def concatenate(arrays, axis=0, out=None):
    """
    Join a sequence of instances of ``BathArray`` along an existing axis. Overrides ``numpy.concatenate`` function.

    Args:
        arrays (list of BathArray): Arrays to concatenate.
            The bath must have the same shape, except in the dimension corresponding to axis (the first, by default).
        axis (int): ``axis`` argument of ``numpy.concatenate``. The axis along which the arrays will be joined.
            If axis is None, arrays are flattened before use. Default is 0.
        out (BathArray): ``out`` argument of ``numpy.concatenate``. If provided, the destination to place the result.

    Returns:
       BathArray: Concatenated array.
    """
    new_array = np.concatenate([x.view(np.ndarray) for x in arrays], axis=axis, out=out)
    new_array = new_array.view(BathArray)
    types = SpinDict()
    imap = InteractionMap()

    offset = 0
    state = BathState(new_array.size)

    for x in arrays:
        types += x.types

        if x.imap:
            imap += x.imap.shift(offset, inplace=False)

        state[offset:offset + x.size] = x.state
        offset += x.size

    new_array.types = types
    new_array._state = state

    if imap:
        new_array.imap = imap

    return new_array


# @implements(np.broadcast_to)
# def broadcast_to(array, shape):
#     ...  # implementation of broadcast_to for MyArray objects

def check_gyro(gyro):
    """
    Check if gyro is matrix or scalar.

    Args:
        gyro (ndarray or float): Gyromagnetic ratio matrix or float.

    Returns:
        tuple: tuple containing:

            * **ndarray or float**: Gyromagnetic ratio.
            * **bool**: True if gyro is float, False otherwise.
    """
    try:
        gyro = float(gyro)
        check = True
    except TypeError:
        check = False

    if not check:
        gyro = np.asarray(gyro)
        if gyro.ndim == 1:  # Assume array
            check = True
        elif not gyro.shape or gyro.shape[0] == 1:
            check = True
            gyro = gyro.reshape(1)[0]
        else:
            test_gyros = gyro.copy()
            indexes = np.arange(gyro.shape[-1])
            test_gyros[..., indexes, indexes] = 0

            diag_check = np.isclose(test_gyros, 0).all()
            same_check = ((gyro[..., 0, 0] == gyro[..., 1, 1]) & (gyro[..., 1, 1] == gyro[..., 2, 2])).all()
            check = diag_check & same_check
            if check:
                gyro = gyro[..., 0, 0][()]

    return gyro, check


def point_dipole(pos, gyro_array, gyro_center):
    """
    Generate an array hyperfine couplings, assuming point dipole approximation.

    Args:
        pos (ndarray with shape (n, 3)): Relative position of the bath spins.
        gyro_array (ndarray with shape (n,)): Array of the gyromagnetic ratios of the bath spins.

        gyro_center (float or ndarray with shape (3, 3)):
            gyromagnetic ratio of the central spin

            **OR**

            tensor corresponding to interaction between magnetic field and central spin.

    Returns:
        ndarray with shape (n, 3, 3): Array of hyperfine tensors.
    """
    identity = np.eye(3, dtype=np.float64)

    try:
        posxpos = np.einsum('...i,...j->...ij', pos, pos)
    except ValueError:
        posxpos = np.outer(pos, pos)

    r = np.linalg.norm(pos, axis=-1)[..., np.newaxis, np.newaxis]
    gyro_center, check = check_gyro(gyro_center)
    gyro_array, check_array = check_gyro(gyro_array)

    gyro_center = np.asarray(gyro_center)
    gyro_array = np.asarray(gyro_array)

    if check and check_array:

        if gyro_center.shape and gyro_array.shape:
            pref = gyro_center[:, np.newaxis] * gyro_array[np.newaxis, :]

        else:
            pref = gyro_center * gyro_array
        out = -(3 * posxpos - identity * r ** 2) / (r ** 5)
        out = out * pref.reshape(*pref.shape, 1, 1) * HBAR_MU0_O4PI / PI2
        return out

    out = -(3 * posxpos - identity * r ** 2) / (r ** 5) * HBAR_MU0_O4PI / PI2

    if not check_array:
        out = np.matmul(out, gyro_array)
    else:
        out *= gyro_array.reshape(*gyro_array.shape, 1, 1)
    if not check:
        out = np.matmul(gyro_center, out)
    else:
        out *= gyro_center.reshape(*gyro_center.shape, 1, 1)
    return out


def same_bath_indexes(barray_1, barray_2, error_range=0.2, ignore_isotopes=True):
    """
    Find indexes of the same array elements in two ``BathArray`` instances.

    Args:
        barray_1 (BathArray): First array.
        barray_2 (BathArray): Second array.
        error_range (float): If distance between positions in two arrays is smaller than ``error_range``
            they are assumed to be the same.
        ignore_isotopes (bool): True if ignore numbers in the name of the spins. Default True.

    Returns:
        tuple: tuple containing:

            * **ndarray**: Indexes of the elements in the first array found in the second.
            * **ndarray**: Indexes of the elements in the second array found in the first.

    """
    # counter_ext = 0
    dist_matrix = np.linalg.norm(barray_1.xyz[:, np.newaxis, :] - barray_2.xyz[np.newaxis, :, :], axis=-1)

    if ignore_isotopes:
        tb_names = np.core.defchararray.strip(barray_1.N, '1234567890')
        ab_names = np.core.defchararray.strip(barray_2.N, '1234567890')

        same_names = tb_names[:, np.newaxis] == ab_names[np.newaxis, :]

    else:
        same_names = barray_1.N[:, np.newaxis] == barray_2.N[np.newaxis, :]

    criteria = np.logical_and(dist_matrix < error_range, same_names)
    indexes, ext_indexes = np.nonzero(criteria)

    # Check for uniqueness. If several follow the criteria, use the first one appearing.
    _, uind = np.unique(indexes, return_index=True)
    indexes = indexes[uind]
    ext_indexes = ext_indexes[uind]

    return indexes, ext_indexes


def update_bath(total_bath, added_bath, error_range=0.2, ignore_isotopes=True, inplace=True):
    """
    Update the ``BathArray`` with elements from the second array.

    Args:

        total_bath (BathArray): Array which contains elements that will be substituted.
        added_bath (BathArray): Array with elements that will be used to substitute elements in ``total_bath``.

        error_range (float): If distance between positions in two arrays is smaller than ``error_range``
            they are assumed to be the same.

        ignore_isotopes (bool): True if ignore numbers in the name of the spins. Default True.

        inplace (bool): True if changes parameters of the ``total_bath`` array in place.
            If False, returns copy of the array.

    Returns:
        BathArray: updated BathArray instance.
    """
    if not inplace:
        total_bath = total_bath.copy()

    indexes, ext_indexes = same_bath_indexes(total_bath, added_bath, error_range, ignore_isotopes)
    neq = total_bath.nc != added_bath.nc
    if neq and added_bath.nc != 1:
        raise ValueError('Arrays correspond to different number of central spins.')

    for n in added_bath.dtype.names:
        if ignore_isotopes and n == 'N':
            continue
        if n == 'A' and neq:
            total_bath[n][indexes, 0] = added_bath[n][ext_indexes]
        else:
            total_bath[n][indexes] = added_bath[n][ext_indexes]

    return total_bath


def transform(atoms, center=None, cell=None, rotation_matrix=None, style='col', inplace=True):
    """
    Coordinate transformation of BathArray.

    Args:
        atoms (BathArray): Array to be transformed.

        center (ndarray with shape (3,)): (0, 0, 0) position of new coordinates in the initial frame.

        cell (ndarray with shape (3, 3)): Cell vectors in cartesian coordinates
            if initial coordinates of the ``atoms`` are in crystallographic reference frame.

        rotation_matrix (ndarray with shape (3, 3)): Rotation matrix R of the **coordinate system**.

            E.g. ``R @ [0, 0, 1] = [a, b, c]`` where ``[a, b, c]`` are coordinates of the z axis of the new coordinate
            system in the old coordinate system.

            Note, that rotaton is applied after transition from cell coordinates to the cartesian coordinates,
            in which cell vectors are stored.

        style (str): Can have two values: 'col' or 'row'.
            Shows how ``cell`` and ``rotation_matrix`` matrices are given:

                * if 'col', each column of the matrix is a vector in previous coordinates;
                * if 'row' - each row is a new vector.

            Default 'col'.

        inplace (bool): If true, makes inplace changes to the provided array.

    Returns:
        BathArray: Transformed array with bath spins.
    """

    styles = ['col', 'row']
    if style.lower() not in styles:
        raise ValueError('Unsupported style of matrices. Available styles are: ' + ', '.join(*styles))

    if not inplace:
        atoms = atoms.copy()

    if center is None:
        center = np.zeros(3)

    if not atoms.dtype.names:
        atoms -= np.asarray(center)
        atoms = rotate_coordinates(atoms, rotation=rotation_matrix, cell=cell, style=style)

        return atoms

    atoms['xyz'] -= np.asarray(center)

    atoms['xyz'] = rotate_coordinates(atoms['xyz'], rotation=rotation_matrix, cell=cell, style=style)

    if 'A' in atoms.dtype.names:
        atoms['A'] = rotate_tensor(atoms['A'], rotation=rotation_matrix, style=style)

    if 'Q' in atoms.dtype.names:
        atoms['Q'] = rotate_tensor(atoms['Q'], rotation=rotation_matrix, style=style)

    return atoms


def _set_sd_attribute(array, attribute_name, initial_value):
    value = np.asarray(initial_value)

    if array.shape and value.shape == array.shape:
        keys, ind = np.unique(array['N'], return_index=True)
        values = value[ind]

        for k, v in zip(keys, values):
            _inner_set_attr(array.types, k, attribute_name, v)

        return

    if not array.shape:
        _inner_set_attr(array.types, array, attribute_name, initial_value)

        return

    for k in np.unique(array['N']):
        _inner_set_attr(array.types, k, attribute_name, initial_value)

        return


def _inner_set_attr(types, key, attr, value):
    try:
        setattr(types[key], attr, value)
    except KeyError:
        types[key] = (0, 0, 0)
        setattr(types[key], attr, value)
    return


def _get_sd_attribute(array, attribute_name):
    if not array.shape:
        return getattr(array.types[array], attribute_name)

    if attribute_name == 'h':
        if array.size == 1:
            return getattr(array.types[array[0]], attribute_name)

        unique_names = np.unique(array.N)

        if unique_names.size == 1:
            return getattr(array.types[array[0]], attribute_name)

        raise RuntimeError('Hamiltonian terms can be modified only for single spin type at a time')

    if array.size == 1:
        newarr = np.array([getattr(array.types[array.N[0]], attribute_name)])

        if attribute_name == 'dim':
            newarr = newarr.astype(int)

        return newarr

    unique_names = np.unique(array.N)

    if unique_names.size == 1:
        n = unique_names[0]
        v = getattr(array.types[n], attribute_name)
        if attribute_name == 'gyro' and isinstance(v, np.ndarray):
            values = np.tile(v, reps=(array.size, 1, 1))
        else:
            if attribute_name == 'dim':
                ones = np.ones(array.shape, dtype=int)
            else:
                ones = np.ones(array.shape, dtype=np.float64)
            values = v * ones

    else:

        temp_values = []
        check = True
        for n in unique_names:
            v = getattr(array.types[n], attribute_name)
            if attribute_name == 'gyro' and isinstance(v, np.ndarray):
                check = False
            temp_values.append(v)

        if check:
            if attribute_name == 'dim':
                values = np.empty(array.shape, dtype=np.float64)
            else:
                values = np.empty(array.shape, dtype=int)

            for n, v in zip(unique_names, temp_values):
                values[array.N == n] = v
        else:
            values = np.empty((array.size, 3, 3), dtype=np.float64)
            for i, n in enumerate(unique_names):
                v = temp_values[i]
                if not isinstance(v, np.ndarray):
                    temp_values[i] = np.eye(3, dtype=np.float64) * v
                values[array.N == n] = v
    return values


def broadcast_array(array, root=0):
    """
    Using mpi4py broadcast ``BathArray`` or ``CenterArray`` to all processes.
    Args:
        array (BathArray or CenterArray): Array to broadcast.
        root (int): Rank of the process to broadcast from.

    Returns:
        BathArray or CenterArray: Broadcasted array.
    """
    import mpi4py
    comm = mpi4py.MPI.COMM_WORLD
    rank = comm.Get_rank()

    if rank == root:
        parameters = vars(array)
    else:
        array = None
        parameters = None

    nbath = comm.bcast(array, root)
    nparam = comm.bcast(parameters, root)

    for k in nparam:
        setattr(nbath, k, nparam[k])

    return nbath


class SpinType:
    r"""
    Class which contains properties of each spin type in the bath.

    Args:
        name (str): Name of the bath spin.
        s (float): Total spin of the bath spin.

            Default 0.

        gyro (float): Gyromagnetic ratio in rad * kHz / G.

            Default 0.

        q (float): Quadrupole moment in barn (for s > 1/2).

            Default 0.

        detuning (float): Energy detuning from the zeeman splitting in kHz,
            included as an extra :math:`+\omega \hat S_z` term in the Hamiltonian,
            where :math:`\omega` is the detuning.

            Default 0.


    Attributes:

        name (str): Name of the bath spin.
        s (float): Total spin of the bath spin.
        dim (int): Spin dimensionality = 2s + 1.

        gyro (float): Gyromagnetic ratio in rad/(ms * G).
        q (float): Quadrupole moment in barn (for s > 1/2).
        detuning (float): Energy detuning from the zeeman splitting in kHz.

    """

    def __init__(self, name, s=0., gyro=0., q=0., detuning=0.):

        self.name = name
        self.s = s

        self.gyro = gyro
        self.q = q
        self.detuning = detuning
        self._h = None

    def __eq__(self, obj):
        if not isinstance(obj, SpinType):
            return False

        checks = (self.name == obj.name) & (self.s == obj.s) & (
                self.gyro == obj.gyro) & (self.q == obj.q) & (self.detuning == obj.detuning)

        return checks

    @_add_args(_stevens_str_doc)
    @property
    def h(self):
        if self._h is None:
            self._h = {}
        return self._h

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def s(self):
        return self._s

    @s.setter
    def s(self, s):

        try:
            if isinstance(s, np.ndarray) and s.shape:
                raise TypeError

            self._s = float(s)
            self._dim = int(2 * s + 1 + 1e-8)

        except TypeError:
            self._dim = np.asarray(2 * s + 1 + 1e-8).astype(np.int32)
            self._s = np.asarray(s).astype(np.float64)

    @property
    def gyro(self):
        return self._gyro

    @gyro.setter
    def gyro(self, gyro):
        gyro, check = check_gyro(gyro)
        self._gyro = gyro

    @property
    def q(self):
        return self._q

    @q.setter
    def q(self, q):
        self._q = q

    @property
    def detuning(self):
        return self._detuning

    @detuning.setter
    def detuning(self, d):
        try:
            self._detuning = float(d)
        except TypeError:
            self._detuning = np.asarray(d).astype(np.float64)

    @property
    def dim(self):
        return self._dim

    def __repr__(self):
        try:
            base_message = f'{self.name}: ({self.s:.1f}, {self.gyro:.4f}'
        except TypeError:
            base_message = f'{self.name}: ({self.s}, {self.gyro}'

        if np.asarray(self.q).any():
            try:
                m = f', {self.q:.4f}'
            except TypeError:
                m = f', {self.q}'
            base_message += m

        if np.asarray(self.detuning).any():
            try:
                m = f', {self.detuning:.4f}'
            except TypeError:
                m = f', {self.detuning}'
            base_message += m

        base_message += ')'

        return base_message


class SpinDict(UserDict):
    """
    Wrapper class for dictionary tailored for containing properties of the spin types.
    Can take ``np.void`` or ``BathArray`` instances as keys.
    Every entry is instance of the ``SpinType``.

    Each entry of the ``SpinDict`` can be initianlized as follows:

        * As a Tuple containing name (optional), spin, gyromagnetic ratio, quadrupole constant (optional)
          and detuning (optional).
        * As a ``SpinType`` instance.

    Examples:

        >>> types = SpinDict()
        >>> types['1H'] = ('1H', 1 / 2, 26.7519)
        >>> types['2H'] = 1, 4.1066, 0.00286
        >>> types['3H'] = SpinType('3H', 1 / 2, 28.535, 0)
        >>> print(types)
        SpinDict({'1H': (1H, 0.5, 26.7519, 0.0), '2H': (2H, 1, 4.1066, 0.00286), '3H': (3H, 0.5, 28.535, 0)})

    If ``SpinType`` of the given bath spin is not provided, when requested
    ``SpinDict`` will try to find information about the bath spins in the ``common_isotopes``.

    If found, adds an entry to the given ``SpinDict`` instance and
    returns it. Otherwise ``KeyError`` is raised.

    To initiallize several ``SpinType`` entries one can use ``add_types`` method.

    Args:
        *args: Any numbers of arguments which could initialize ``SpinType`` instances.
        **kwargs: Any numbers of keyword arguments which could initialize ``SpinType`` instances.
            For details see ``SpinDict.add_type`` method.

    """

    def __init__(self, *args, **kwargs):
        super(SpinDict, self).__init__()
        self.add_type(*args, **kwargs)

    def __delitem__(self, key):
        try:
            super().__delitem__(key)
        except TypeError:
            if key.shape:
                try:
                    names = key['N']
                except IndexError:
                    names = key
                for k in names:
                    super().__delitem__(k)
                return

            k = key[()]
            try:
                k = k['N']
            except TypeError:
                pass
            super().__delitem__(k)

    def __setitem__(self, key, value):
        try:
            value = _check_key_spintype(key, value)
            super().__setitem__(key, value)

        except (TypeError, ValueError):
            key = np.asarray(key)
            if key.shape:
                try:
                    names = key['N']
                except IndexError:
                    names = key

                for k, v in zip(names, value):
                    v = _check_key_spintype(k, v)
                    super().__setitem__(k, v)
                return

            k = key[()]

            try:
                k = k['N']
            except TypeError:
                pass

            value = _check_key_spintype(k, value)
            super().__setitem__(k, value)

    def __getitem__(self, key):

        try:
            key = key[()]
            return self._super_get_item(key['N'])
            # self._super_get_item(key)

        except (TypeError, IndexError):
            try:
                return self._super_get_item(key)
            except TypeError:

                if key.dtype.names:
                    key = key['N']

                key = np.asarray(key)
                if key.size == 1:
                    n = key[0]
                    spins = np.array([self._super_get_item(n).s])
                    gyros = np.array([self._super_get_item(n).gyro])
                    quads = np.array([self._super_get_item(n).q])
                    detus = np.array([self._super_get_item(n).detuning])
                    return SpinType(key, s=spins, gyro=gyros, q=quads, detuning=detus)
                else:
                    raise TypeError('Unsupported key.')


    # adding two objects
    def __add__(self, obj):
        new_obj = SpinDict()
        keys_1 = list(self.keys())
        keys_2 = list(obj.keys())

        for k in {*keys_1, *keys_2}:

            if (k in keys_1) and (k in keys_2):
                assert obj[k] == self[k], f'Error, type {k} has different properties in provided types'
                new_obj[k] = self[k]

            elif k in keys_1:
                new_obj[k] = self[k]
            else:
                new_obj[k] = obj[k]
        return new_obj

    def __repr__(self):
        message = f"{type(self).__name__}("
        for k in self.data:
            message += f"{self.data[k]}, "
            if len(message) > 75:
                message += '..., '
                break

        message = message[:-2] + ')'

        return message

    def _super_get_item(self, n):
        try:
            return super().__getitem__(n)
        except KeyError:
            if not n in common_isotopes:
                raise KeyError(_spin_not_found_message(n))
            super().__setitem__(n, copy.copy(common_isotopes[n]))
            return super().__getitem__(n)

    def add_type(self, *args, **kwargs):
        """
        Add one or several spin types to the spin dictionary.

        Args:
            *args:
                Any numbers of arguments which could initialize ``SpinType`` instances.
                Accepted arguments:

                    * Tuple containing name, spin, gyromagnetic ratio, quadrupole constant (optional)
                      and detuning (optional).
                    * ``SpinType`` instance.

                Can also initialize one instance of ``SpinType`` if each argument corresponds to
                each positional argument necessary to initiallize.

            **kwargs: Any numbers of keyword arguments which could initialize ``SpinType`` instances.
                Usefull as an alternative for updating the dictionary. for each keyword argument adds an entry
                to the ``SpinDict`` with the same name as keyword.

        Examples:
            >>> types = SpinDict()
            >>> types.add_type('1H', 1 / 2, 26.7519)
            >>> types.add_type(('1H_det', 1 / 2, 26.7519, 10), ('2H', 1, 4.1066, 0.00286),
            >>>                 SpinType('3H', 1 / 2, 28.535, 0), e=(1 / 2, 6.7283, 0))
            >>> print(types)
            SpinDict(1H: (1H, 0.5, 26.7519), 1H_det: (1H_det, 0.5, 26.7519, 10),
            2H: (2H, 1, 4.1066, 0.00286), 3H: (3H, 0.5, 28.535), e: (e, 0.5, 6.7283))

        """
        keys = []
        try:
            st = SpinType(*args, **kwargs)
            self[st.name] = st
            return
        except TypeError:
            try:
                for nuc in args:
                    if isinstance(nuc, SpinType):
                        key = nuc.name
                        self[key] = nuc
                        keys.append(key)
                    elif isinstance(nuc, Mapping):
                        self.update(nuc)
                        keys += list(nuc.keys())
                    else:
                        key = nuc[0]
                        self[key] = SpinType(*nuc)
                        keys.append(key)
            except TypeError:
                for k in keys:
                    self.pop(k)
                self[args[0]] = SpinType(*args)
            for nuc in kwargs:
                self[nuc] = kwargs[nuc]


def _check_key_spintype(k, v):
    if isinstance(v, SpinType):
        return v

    if v[0] == k:
        v = SpinType(*v)
    else:
        v = SpinType(k, *v)
    return v


_spin_not_found_message = lambda x: 'Spin type for {} was not provided and was not found in common isotopes.'.format(x)

import pandas as pd

# try:
#     url = 'https://raw.githubusercontent.com/StollLab/EasySpin/main/easyspin/private/isotopedata.txt'
#     all_spins = pd.read_csv(url, delim_whitespace=True, header=None, comment='%',
#                             names=['protons', 'nucleons', 'radioactive', 'symbol', 'name', 'spin', 'g', 'conc', 'q'])
# except:
import os

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
filepath = os.path.join(__location__, 'isotopes.txt')
all_spins = pd.read_csv(filepath, delim_whitespace=True, header=None, comment='%',
                        names=['protons', 'nucleons', 'radioactive', 'symbol', 'name', 'spin', 'g', 'conc', 'q'])

# only stable isotopes with nonzero spins
spins = all_spins[(all_spins['spin'] > 0) & (all_spins['conc'] > 0)]

_names = spins['nucleons'].astype(str) + spins['symbol']
_gyros = spins['g'] / HBAR_SI * NUCLEAR_MAGNETON / 1e7
_quads = spins['q']
_spins = spins['spin']

_mi = pd.MultiIndex.from_arrays([spins['symbol'], _names])
_ser = pd.Series((spins['conc'] / 100).values, index=_mi)

common_concentrations = {level: _ser.xs(level).to_dict() for level in _ser.index.levels[0]}
"""
dict: Nested dict containing natural concentrations of the stable nuclear isotopes.  
"""

# Dictionary of the common isotopes. Placed in this file to avoid circular dependency
common_isotopes = SpinDict(*zip(_names, _spins, _gyros, _quads))
"""
SpinDict: An instance of the ``SpinDict`` dictionary, containing properties for the most of the common isotopes with 
nonzero spin.
The isotope is considered common if it is stable and has nonzero concentration in nature.  
"""

# electron spin
common_isotopes['e'] = SpinType('e', 1 / 2, ELECTRON_GYRO, 0)
# allias for common_isotopes
ci = common_isotopes



import warnings

import numpy as np
from pycce.bath.array import BathArray, transform
from pycce.constants import MHZ_TO_KHZ, BOHR_TO_ANGSTROM, EFG_CONVERSION
from pycce.io.base import DFTCoordinates, fortran_value, find_first_index, set_isotopes

qe_coord_types = ['crystal', 'bohr', 'angstrom', 'alat']


def read_qe(pwfile, hyperfine=None, efg=None, s=1, pwtype=None, types=None, isotopes=None,
            center=None, center_type=None, rotation_matrix=None, rm_style='col', find_isotopes=True):
    r"""
    Function to read PW/GIPAW output from Quantum Espresso into BathArray.

    Changes the names of the atoms to the most abundant isotopes if ``find_isotopes`` set to True.
    If that is not the desired outcome, user can define which isotopes to use using keyword isotopes.
    If ``find_isotopes`` is False, then keep the original names even when ``isotopes`` argument is provided.

    Args:
        pwfile (str):
            Name of PW input or output file.
            If the file doesn't have proper extension, parameter pw_type should indicate the type.
        hyperfine (str): name of the GIPAW hyperfine output.
        efg (str): Name of the gipaw electric field tensor output.
        s (float): Spin of the central spin. Default 1.
        pwtype (str): Type of the ``pwfile``. if not listed, will be inferred from extension of pwfile.
        types (SpinDict or list of tuples): SpinDict containing SpinTypes of isotopes or input to make one.
        isotopes (dict): Optional.
            Dictionary with entries: {"element" : "isotope"}, where "element" is the name of the element
            in DFT output, "isotope" is the name of the isotope.
        center (ndarray of shape (3,)): Position of (0, 0, 0) point in input coordinates.
        center_type (str): Type of the coordinates provided in center argument.
            Possible value include: 'bohr', 'angstrom', 'crystal', 'alat'. Default assumes the same as in PW file.
        rotation_matrix (ndarray of shape (3,3)):
            Rotation matrix to rotate basis. For details see utilities.transform.
        rm_style (str):
            Indicates how rotation matrix should be interpreted.
            Can take values "col" or "row". Default "col"
        find_isotopes (bool): If true, sets isotopes instead of names of the atoms.

    Returns:
        BathArray:
            BathArray containing atoms with hyperfine couplings and quadrupole tensors from QE output.

    """

    couplings = None
    gradients = None

    if hyperfine is not None:
        contact, dipolar = read_hyperfine(hyperfine, spin=s)
        couplings = dipolar + np.eye(3)[np.newaxis, :, :] * contact[:, np.newaxis, np.newaxis]

    if efg is not None:
        gradients = read_efg(efg)

    pwoutput = PWCoordinates(pwfile, pwtype=pwtype)

    if center is not None and center_type is None:
        center_type = pwoutput.coordinates_units

    pwoutput.to_angstrom(inplace=True)
    coord, names = pwoutput.coordinates, pwoutput.names

    with warnings.catch_warnings(record=True) as w:
        atoms = BathArray(array=coord, names=names,
                          hyperfines=couplings,
                          types=types)

    if find_isotopes:
        set_isotopes(atoms, isotopes=isotopes, spin_types=types)

    if gradients is not None:
        atoms.from_efg(gradients)

    if center is not None:
        center = pwoutput.get_angstrom(center, center_type)

    atoms = transform(atoms, center, rotation_matrix=rotation_matrix, style=rm_style)
    return atoms


class PWCoordinates(DFTCoordinates):
    """
    Coordinates of the system from the PW data of Quantum Espresso. Subclass of the DFTCoordinates.

    With initiallization reads either output or input of PW module of QE.

    Args:
        filename (str): name of the PW input or output.
        pwfile (str):
            Name of PW input or output file.
            If the file doesn't have proper extension, parameter pw_type should indicate the type.
        pwtype (str): Type of the coord_f. if not listed, will be inferred from extension of pwfile.
        to_angstrom (bool): True if automatically convert the units of ``cell`` and ``coordinates`` to Angstrom.

    """

    def __init__(self, filename, pwtype=None, to_angstrom=False):
        super().__init__()

        if not pwtype:
            pwtype = filename.split('.')[-1]

        if pwtype == 'in':
            self.parse_input(filename, to_angstrom=to_angstrom)
        elif pwtype == 'out':
            self.parse_output(filename, to_angstrom=to_angstrom)

        else:
            raise TypeError('Unsupported pw_type! Only .out or .in are supported')

    def parse_output(self, filename, to_angstrom=False):
        """
        Method to read coordinates of atoms from PW output into the PWCoordinates instance.

        Args:
            filename (str): the name of the output file.
            to_angstrom (bool): True if automatically convert the units of ``cell`` and ``coordinates`` to Angstrom.

        Returns:
            None

        """
        with open(filename) as f:
            lines = f.readlines()
        alat_index = find_first_index('lattice parameter (alat)', lines)

        self.alat = float(lines[alat_index].split()[-2]) * BOHR_TO_ANGSTROM

        cell_index = find_first_index('crystal axes', lines) + 1
        cell = []
        for index in range(cell_index, cell_index + 3):
            row_split = lines[index].split()
            cell_row = [float(x) for x in row_split[-4:-1]]
            cell.append(cell_row)

        self.cell = np.asarray(cell).T
        self.cell_units = 'alat'

        # read coordinates

        names = []
        coordinates = []
        try:
            final_kw = 'Begin final coordinates'
            start = find_first_index(final_kw, lines)
            coord_kw = 'ATOMIC_POSITIONS'
            index = find_first_index(coord_kw, lines, start=start)
            coord_units = get_ctype(lines[index])

            while True:
                try:
                    index += 1

                    row_split = lines[index].split()
                    name = row_split[0]
                    crow = [float(x) for x in row_split[1:]]

                    names.append(name)
                    coordinates.append(crow)
                except (IndexError, ValueError):
                    break

        except KeyError:
            coord_units = 'alat'

            coord_kw = 'Cartesian axes'
            index = find_first_index(coord_kw, lines) + 2
            while True:
                try:
                    index += 1

                    row_split = lines[index].split()
                    name = row_split[1]
                    crow = [float(x) for x in row_split[-4:-1]]

                    names.append(name)
                    coordinates.append(crow)

                except IndexError:
                    break

        self.coordinates_units = coord_units
        self.coordinates = np.asarray(coordinates)
        self.names = np.array(names, dtype='<U16')

        if to_angstrom:
            self.to_angstrom(inplace=True)
        return self

    def parse_input(self, filename, to_angstrom=False):
        """
        Method to read coordinates of atoms from PW input into the PWCoordinates instance.

        Args:
            filename (str): the name of the output file.
            to_angstrom (bool): True if automatically convert the units of ``cell`` and ``coordinates`` to Angstrom.

        """
        with open(filename) as f:
            input_string = f.read()

        namelists = read_qe_namelists(input_string.lower())
        lines = input_string.splitlines()

        alat = namelists['system'].get('celldm(1)', None)

        if alat is None:
            alat = namelists['system'].get('a', None)
        else:
            alat *= BOHR_TO_ANGSTROM

        cell = cell_from_system(namelists['system'])
        if cell is not None:
            cell_units = 'bohr'
        else:
            index = find_first_index('CELL_PARAMETERS', lines)
            cell_units = get_ctype(lines[index])
            self.cell_units = cell_units
            cell = []

            for _ in range(3):
                index += 1
                line = lines[index]
                cell.append([float(x) for x in line.split()])

            cell = np.asarray(cell).T

            if cell_units == 'alat' and alat is None:
                raise ValueError('alat was not found')

            if alat is None:
                assert cell_units in ['bohr', 'angstrom']
                alat = cell[0, 0] * BOHR_TO_ANGSTROM ** (cell_units == 'bohr')

        self.cell = cell
        self.cell_units = cell_units
        self.alat = alat

        index = find_first_index('ATOMIC_POSITIONS', lines)
        coord_units = get_ctype(lines[index])

        names = []
        coords = []

        for i in range(index + 1, index + 1 + namelists['system']['nat']):
            row_split = lines[i].split()
            names.append(row_split[0])
            coords.append([float(x) for x in row_split[1:]])

        self.names = np.array(names, dtype='<U16')
        self.coordinates = np.asarray(coords)
        self.coordinates_units = coord_units

        if to_angstrom:
            self.to_angstrom(inplace=True)


def cell_from_system(sdict):
    """
    Function to obtain cell from namelist SYSTEM read from PW input.

    Args:
        sdict (dict): Dictinary generated from namelist SYSTEM of PW input.

    Returns:
        ndarray with shape (3,3):
            Cell is 3x3 matrix with entries::

                [[a_x b_x c_x]
                 [a_y b_y c_y]
                 [a_z b_z c_z]],

            where a, b, c are crystallographic vectors,
            and x, y, z are their coordinates in the cartesian reference frame.

    """
    ibrav = sdict.get('ibrav', None)
    if ibrav == 0:
        return None
    params = ['a', 'b', 'c', 'cosab', 'cosac', 'cosbc']
    celldm = [sdict.get(f'celldm({i + 1})', 0) for i in range(6)]
    if not any(celldm):
        abc = [sdict.get(a, 0) for a in params]
        celldm = celldms_from_abc(ibrav, abc)

    if not any(celldm):
        return None

    if ibrav == 1:
        cell = np.eye(3) * celldm[0]
        return cell

    elif ibrav == 2:
        v1 = celldm[0] / 2 * np.array([-1, 0, 1])
        v2 = celldm[0] / 2 * np.array([0, 1, 1])
        v3 = celldm[0] / 2 * np.array([-1, 1, 0])

    elif ibrav == 3:
        v1 = celldm[0] / 2 * np.array([1, 1, 1])
        v2 = celldm[0] / 2 * np.array([-1, 1, 1])
        v3 = celldm[0] / 2 * np.array([-1, -1, 1])

    elif ibrav == -3:
        v1 = celldm[0] / 2 * np.array([-1, 1, 1])
        v2 = celldm[0] / 2 * np.array([1, -1, 1])
        v3 = celldm[0] / 2 * np.array([1, 1, -1])

    elif ibrav == 4:
        v1 = celldm[0] * np.array([1, 0, 0])
        v2 = celldm[0] * np.array([-1 / 2, np.sqrt(3) / 2, 0])
        v3 = celldm[0] * np.array([0, 0, celldm[2]])

    elif ibrav == 5:
        term_1 = np.sqrt(1 + 2 * celldm[3])
        term_2 = np.sqrt(1 - celldm[3])
        v1 = celldm[0] * np.array([term_2 / np.sqrt(2), -term_2 / np.sqrt(6), term_1 / np.sqrt(3)])
        v2 = celldm[0] * np.array([0, term_2 * np.sqrt(2 / 3), term_1 / np.sqrt(3)])
        v3 = celldm[0] * np.array([-term_2 / np.sqrt(2), -term_2 / np.sqrt(6), term_1 / np.sqrt(3)])

    elif ibrav == -5:
        term_1 = np.sqrt(1 + 2 * celldm[3])
        term_2 = np.sqrt(1 - celldm[3])
        v1 = celldm[0] * np.array([(term_1 - 2 * term_2) / 3, (term_1 + term_2) / 3, (term_1 + term_2) / 3])
        v2 = celldm[0] * np.array([(term_1 + term_2) / 3, (term_1 - 2 * term_2) / 3, (term_1 + term_2) / 3])
        v3 = celldm[0] * np.array([(term_1 + term_2) / 3, (term_1 + term_2) / 3, (term_1 - 2 * term_2) / 3])

    elif ibrav == 6:
        v1 = celldm[0] * np.array([1, 0, 0])
        v2 = celldm[0] * np.array([0, 1, 0])
        v3 = celldm[0] * np.array([0, 0, celldm[2]])
    elif ibrav == 7:
        v1 = celldm[0] / 2 * np.array([1, -1, celldm[2]])
        v2 = celldm[0] / 2 * np.array([1, 1, celldm[2]])
        v3 = celldm[0] / 2 * np.array([-1, -1, celldm[2]])
    elif ibrav == 8:
        v1 = celldm[0] * np.array([1, 0, 0])
        v2 = celldm[0] * np.array([0, celldm[1], 0])
        v3 = celldm[0] * np.array([0, 0, celldm[2]])
    elif ibrav == 9:
        v1 = celldm[0] / 2 * np.array([1, celldm[1], 0])
        v2 = celldm[0] / 2 * np.array([-1, celldm[1], 0])
        v3 = celldm[0] * np.array([0, 0, celldm[2]])

    elif ibrav == -9:
        v1 = celldm[0] / 2 * np.array([1, -celldm[1], 0])
        v2 = celldm[0] / 2 * np.array([+1, celldm[1], 0])
        v3 = celldm[0] * np.array([0, 0, celldm[2]])

    elif ibrav == 91:
        v1 = celldm[0] * np.array([1, 0, 0])
        v2 = celldm[0] / 2 * np.array([0, celldm[1], -celldm[2]])
        v3 = celldm[0] / 2 * np.array([0, celldm[1], celldm[2]])
    elif ibrav == 10:
        v1 = celldm[0] / 2 * np.array([1, 0, celldm[2]])
        v2 = celldm[0] / 2 * np.array([1, celldm[1], 0])
        v3 = celldm[0] / 2 * np.array([0, celldm[1], celldm[2]])
    elif ibrav == 11:
        v1 = celldm[0] / 2 * np.array([1, celldm[1], celldm[2]])
        v2 = celldm[0] / 2 * np.array([-1, celldm[1], celldm[2]])
        v3 = celldm[0] / 2 * np.array([-1, -celldm[1], celldm[2]])
    elif ibrav == 12:
        sen = np.sqrt(1 - celldm[3] ** 2)
        v1 = celldm[0] * np.array([1, 0, 0])
        v2 = celldm[0] * np.array([celldm[1] * celldm[3], celldm[1] * sen, 0])
        v3 = celldm[0] * np.array([0, 0, celldm[2]])
    elif ibrav == -12:
        sen = np.sqrt(1 - celldm[4] ** 2)
        v1 = celldm[0] * np.array([1, 0, 0])
        v2 = celldm[0] * np.array([0, celldm[1], 0])
        v3 = celldm[0] * np.array([celldm[2] * celldm[4], 0, celldm[2] * sen])

    elif ibrav == 13:
        sen = np.sqrt(1 - celldm[3] ** 2)
        v1 = celldm[0] / 2 * np.array([1, 0, -celldm[2]])
        v2 = celldm[0] * np.array([celldm[1] * celldm[3], celldm[1] * sen, 0])
        v3 = celldm[0] / 2 * np.array([1, 0, celldm[2]])

    elif ibrav == -13:
        sen = np.sqrt(1 - celldm[4] ** 2)
        v1 = celldm[0] / 2 * np.array([1, celldm[1], 0])
        v2 = celldm[0] / 2 * np.array([-1, celldm[1], 0])
        v3 = celldm[0] * np.array([celldm[2] * celldm[4], 0, celldm[2] * sen])

    elif ibrav == 14:
        singam = np.sqrt(1 - celldm[5] ** 2)
        term = (1 + 2 * celldm[3] * celldm[4] * celldm[5] - celldm[3] ** 2 - celldm[4] ** 2 - celldm[5] ** 2)
        term = np.sqrt(term / (1 - celldm[5] ** 2))

        v1 = celldm[0] * np.array([1,
                                   0,
                                   0])
        v2 = celldm[0] * np.array([celldm[1] * celldm[5],
                                   celldm[1] * singam,
                                   0])
        v3 = celldm[0] * np.array([celldm[2] * celldm[4],
                                   celldm[2] * (celldm[3] - celldm[4] * celldm[5]) / singam,
                                   celldm[2] * term])
    else:
        raise ValueError('Unsupported ibrav')

    cell = np.stack([v1, v2, v3], axis=1)

    return cell


def celldms_from_abc(ibrav, abc_list):
    """
    Obtain celldms from ibrav value and a, b, c, cosab, cosac, cosbc parameters.

    Using ibrav value and abc parameters from PW input generate celldm array, necessary to construct cell parameters.
    For details about abc and ibrav values see PW input documentation.

    Args:
        ibrav (int): ibrav parameter of PW input.
        abc_list (list): List, of 6 parameters:  a, b, c, cosab, cosac, cosbc

    Returns:
        celldm (list): list of 6 values, from which cell can be generated.

    """
    a, b, c, cosab, cosac, cosbc = abc_list

    celldm = [0.] * 6
    celldm[0] = a / BOHR_TO_ANGSTROM
    celldm[1] = b / a
    celldm[2] = c / a

    if ibrav in [0, 14]:

        celldm[3] = cosbc
        celldm[4] = cosac
        celldm[5] = cosab

    elif ibrav in [-12, -13]:
        celldm[3] = 0.0
        celldm[4] = cosac
        celldm[5] = 0.0

    elif ibrav in [-5, 5, 12, 13]:
        celldm[3] = cosab
        celldm[4] = 0.0
        celldm[5] = 0.0

    return celldm


def read_gipaw_tensors(lines, keyword=None, start=None, conversion=1):
    """
    Helper function to read GIPAW tensors from the list of lines.

    Args:
        lines (list of str): List of strings contraining lines from the file. Output of open(file).readlines().
        keyword (str): Keyword in the line which indicates the beginning of the tensor data block.
        start (int): Index of the line which indicates the beginning of the tensor data block.
        conversion (float): Conversion factor from GIPAW units to the ones, used in this package.

    Returns:
        ndarray with shape (n, 3, 3): Array of tensors.

    """
    if keyword is not None:
        start = find_first_index(keyword, lines)

    if start is None:
        raise ValueError

    all_tensors = []

    n = start
    while True:
        n += 1
        tensor = []
        try:
            for _ in range(3):
                line = lines[n]
                tensor.append([float(x) * conversion for x in line.split()[2:]])
                n += 1

        except (ValueError, IndexError):
            break

        all_tensors.append(tensor)

    return all_tensors


def read_hyperfine(filename, spin=1):
    """
    Function to read hyperfine couplings from GIPAW output.

    Args:
        filename (str): Name of the GIPAW hyperfine output.
        spin (float): Spin of the central spin. Default 1.

    Returns:
        tuple: Tuple containing:

            * *ndarray with shape (n,)*: Array of Fermi contact terms.
            * *ndarray with shape (n, 3,3)*: Array of spin dipolar hyperfine tensors.

    """
    conversion = MHZ_TO_KHZ / (2 * spin)

    with open(filename) as f:
        lines = f.readlines()

    dipol_keyword = 'total dipolar (symmetrized)'
    contact_keyword = 'Fermi contact in MHz'

    dipolars = read_gipaw_tensors(lines, dipol_keyword, conversion=conversion)
    start = find_first_index(contact_keyword, lines) + 2

    contacts = []

    for index in range(start, start + len(dipolars)):
        line = lines[index]
        # divided by spin, b/c NI in the GIPAW code
        cont = float(line.split()[-1]) * conversion
        contacts.append(cont)

    return np.asarray(contacts), np.asarray(dipolars)


def read_efg(filename):
    """
    Function to read electric field gradient tensors from GIPAW output.

    Args:
        filename (str): Name of the GIPAW EFG-containing output.

    Returns:
        ndarray with shape (n, 3,3): Array of EFG tensors.

    """
    efg_kw = 'total EFG (symmetrized)'
    with open(filename) as f:
        lines = f.readlines()

    tensors = read_gipaw_tensors(lines, keyword=efg_kw, conversion=EFG_CONVERSION)

    return np.asarray(tensors)


def read_qe_namelists(input_string):
    """
    Read Fortran-like namelists from the large string.

    Args:
        input_string (str): String representation of the QE input file.

    Returns:
        dict: Dictionary, containing dicts for each namelist found in the input string.

    """
    namelists = {}

    for block in input_string.split('/\n'):

        if not '&' in block:
            # means we are out of namelists
            break

        lines = [s.strip() for s in block.splitlines() if (s.strip() and s[0] != '!')]
        index = find_first_index('&', lines)
        block_name = lines[index].strip('&\t ')

        namelists[block_name] = {}
        for row in lines[index + 1:]:
            for pair in row.split(','):
                if not pair.strip():
                    continue
                name, _, value = (x.strip(',\t ').lower() for x in pair.partition("="))
                namelists[block_name][name] = fortran_value(value)

    return namelists


def get_ctype(lin):
    """
    Get coordinates type from the line of QE input/output.

    Args:
        str: Line from QE input/output containing string with coordinates type.

    Returns:
        str: type of the coordinates.
    """

    try:
        coord_type = next(filter(lambda x: x in lin.lower(), qe_coord_types))
    except IndexError:
        raise ValueError(f'{lin} type is not supported.\nAllowed types: ', ' '.join(*qe_coord_types))

    return coord_type

# ELSEIF (ibrav == 7) THEN
#    !
#    !     body centered tetragonal lattice
#    !
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    !
#    cbya=celldm(3)
#    a2(1)=celldm(1)/2.d0
#    a2(2)=a2(1)
#    a2(3)=cbya*celldm(1)/2.d0
#    a1(1)= a2(1)
#    a1(2)=-a2(1)
#    a1(3)= a2(3)
#    a3(1)=-a2(1)
#    a3(2)=-a2(1)
#    a3(3)= a2(3)
#    !
# ELSEIF (ibrav == 8) THEN
#    !
#    !     Simple orthorhombic lattice
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    !
#    a1(1)=celldm(1)
#    a2(2)=celldm(1)*celldm(2)
#    a3(3)=celldm(1)*celldm(3)
#    !
# ELSEIF ( abs(ibrav) == 9) THEN
#    !
#    !     One face (base) centered orthorhombic lattice  (C type)
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', &
#                                                                abs(ibrav))
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', &
#                                                                abs(ibrav))
#    !
#    IF ( ibrav == 9 ) THEN
#       !   old PWscf description
#       a1(1) = 0.5d0 * celldm(1)
#       a1(2) = a1(1) * celldm(2)
#       a2(1) = - a1(1)
#       a2(2) = a1(2)
#    ELSE
#       !   alternate description
#       a1(1) = 0.5d0 * celldm(1)
#       a1(2) =-a1(1) * celldm(2)
#       a2(1) = a1(1)
#       a2(2) =-a1(2)
#    ENDIF
#    a3(3) = celldm(1) * celldm(3)
#    !
# ELSEIF ( ibrav == 91 ) THEN
#    !
#    !     One face (base) centered orthorhombic lattice  (A type)
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    !
#    a1(1) = celldm(1)
#    a2(2) = celldm(1) * celldm(2) * 0.5_DP
#    a2(3) = - celldm(1) * celldm(3) * 0.5_DP
#    a3(2) = a2(2)
#    a3(3) = - a2(3)
#    !
# ELSEIF (ibrav == 10) THEN
#    !
#    !     All face centered orthorhombic lattice
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    !
#    a2(1) = 0.5d0 * celldm(1)
#    a2(2) = a2(1) * celldm(2)
#    a1(1) = a2(1)
#    a1(3) = a2(1) * celldm(3)
#    a3(2) = a2(1) * celldm(2)
#    a3(3) = a1(3)
#    !
# ELSEIF (ibrav == 11) THEN
#    !
#    !     Body centered orthorhombic lattice
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    !
#    a1(1) = 0.5d0 * celldm(1)
#    a1(2) = a1(1) * celldm(2)
#    a1(3) = a1(1) * celldm(3)
#    a2(1) = - a1(1)
#    a2(2) = a1(2)
#    a2(3) = a1(3)
#    a3(1) = - a1(1)
#    a3(2) = - a1(2)
#    a3(3) = a1(3)
#    !
# ELSEIF (ibrav == 12) THEN
#    !
#    !     Simple monoclinic lattice, unique (i.e. orthogonal to a) axis: c
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    IF (abs(celldm(4))>=1.d0) CALL errore ('latgen', 'wrong celldm(4)', ibrav)
#    !
#    sen=sqrt(1.d0-celldm(4)**2)
#    a1(1)=celldm(1)
#    a2(1)=celldm(1)*celldm(2)*celldm(4)
#    a2(2)=celldm(1)*celldm(2)*sen
#    a3(3)=celldm(1)*celldm(3)
#    !
# ELSEIF (ibrav ==-12) THEN
#    !
#    !     Simple monoclinic lattice, unique axis: b (more common)
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)',-ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)',-ibrav)
#    IF (abs(celldm(5))>=1.d0) CALL errore ('latgen', 'wrong celldm(5)',-ibrav)
#    !
#    sen=sqrt(1.d0-celldm(5)**2)
#    a1(1)=celldm(1)
#    a2(2)=celldm(1)*celldm(2)
#    a3(1)=celldm(1)*celldm(3)*celldm(5)
#    a3(3)=celldm(1)*celldm(3)*sen
#    !
# ELSEIF (ibrav == 13) THEN
#    !
#    !     One face centered monoclinic lattice unique axis c
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    IF (abs(celldm(4))>=1.d0) CALL errore ('latgen', 'wrong celldm(4)', ibrav)
#    !
#    sen = sqrt( 1.d0 - celldm(4) ** 2 )
#    a1(1) = 0.5d0 * celldm(1)
#    a1(3) =-a1(1) * celldm(3)
#    a2(1) = celldm(1) * celldm(2) * celldm(4)
#    a2(2) = celldm(1) * celldm(2) * sen
#    a3(1) = a1(1)
#    a3(3) =-a1(3)
# ELSEIF (ibrav == -13) THEN
#    !
#    !     One face centered monoclinic lattice unique axis b
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)',-ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)',-ibrav)
#    IF (abs(celldm(5))>=1.d0) CALL errore ('latgen', 'wrong celldm(5)',-ibrav)
#    !
#    sen = sqrt( 1.d0 - celldm(5) ** 2 )
#    a1(1) = 0.5d0 * celldm(1)
#    a1(2) =-a1(1) * celldm(2)
#    a2(1) = a1(1)
#    a2(2) =-a1(2)
#    a3(1) = celldm(1) * celldm(3) * celldm(5)
#    a3(3) = celldm(1) * celldm(3) * sen
#    !
# ELSEIF (ibrav == 14) THEN
#    !
#    !     Triclinic lattice
#    !
#    IF (celldm (2) <= 0.d0) CALL errore ('latgen', 'wrong celldm(2)', ibrav)
#    IF (celldm (3) <= 0.d0) CALL errore ('latgen', 'wrong celldm(3)', ibrav)
#    IF (abs(celldm(4))>=1.d0) CALL errore ('latgen', 'wrong celldm(4)', ibrav)
#    IF (abs(celldm(5))>=1.d0) CALL errore ('latgen', 'wrong celldm(5)', ibrav)
#    IF (abs(celldm(6))>=1.d0) CALL errore ('latgen', 'wrong celldm(6)', ibrav)
#    !
#    singam=sqrt(1.d0-celldm(6)**2)
#    term= (1.d0+2.d0*celldm(4)*celldm(5)*celldm(6)             &
#         -celldm(4)**2-celldm(5)**2-celldm(6)**2)
#    IF (term < 0.d0) CALL errore &
#       ('latgen', 'celldm do not make sense, check your data', ibrav)
#    term= sqrt(term/(1.d0-celldm(6)**2))
#    a1(1)=celldm(1)
#    a2(1)=celldm(1)*celldm(2)*celldm(6)
#    a2(2)=celldm(1)*celldm(2)*singam
#    a3(1)=celldm(1)*celldm(3)*celldm(5)
#    a3(2)=celldm(1)*celldm(3)*(celldm(4)-celldm(5)*celldm(6))/singam
#    a3(3)=celldm(1)*celldm(3)*term



from collections import defaultdict

import numpy as np
import scipy.linalg
from pycce.utilities import expand, shorten_dimensions


class Propagator:
    r"""
    Class containing properties of the Hamiltonian.

    Essentially wrapper for ndarray with additional attributes of ``dimensions`` and ``spins``.

    Usual methods (e.g. ``__setitem__`` or ``__getitem__``) access the ``data`` attribute.

    .. note::

        Algebraic operations with Hamiltonian will return ndarray instance.

    Args:
        dimensions (array-like): array of the dimensions for each spin in the Hilbert space of the Hamiltonian.

    Attributes:
        dimensions (ndarray): array of the dimensions for each spin in the Hilbert space of the Hamiltonian.
        spins (ndarray): array of the spins, spanning the Hilbert space of the Hamiltonian.
        vectors (list): list with spin vectors of form ``[[Ix, Iy, Iz], [Ix, Iy, Iz], ...]``.
        data (ndarray): matrix representation of the Hamiltonian.

    """

    def __init__(self, base_hamiltonian):
        self.base_hamiltonian = base_hamiltonian

    def generate_base_hamiltonian(self):
        return 1

    def generate_rotations(self, pulses, dimensions=None,
                           bath=None, vectors=None, central_spin=None):
        """
        Generate list of matrix representations of the rotations, induced by the sequence of the pulses.

        The rotations are stored in the ``.rotation`` attribute of the each ``Pulse`` object
        and in ``Sequence.rotations``.

        Args:
            dimensions (ndarray with shape (N,)): Array of spin dimensions in the system.
            bath (BathArray with shape (n,)): Array of bath spins in the system.
            vectors (ndarray  with shape (N, 3, prod(dimensions), prod(dimensions))):
                 Array with vectors of spin matrices for each spin in the system.

            central_spin (CenterArray): Optional. Array of central spins.

        Returns:
            tuple: *tuple* containing:

            * **list** or **None**: List with delays before each pulse or None if equispaced.

            * **list**: List with matrix representations of the rotation from each pulse.
        """
        cs = central_spin
        ndims = len(dimensions)
        if cs is not None:
            nc = len(cs)  # number of central spins
            shortdims = shorten_dimensions(dimensions, nc)

        self.delays = None
        self.rotations = None
        equispaced = not any(p._has_delay for p in pulses.data)

        if equispaced:
            delays = None

        else:
            delays = [p.delay if p.delay is not None else 0 for p in pulses.data]

        rots = []
        # Sigma as in central spin array is total spin
        total_sigma = {}
        separate_sigma = defaultdict(dict)
        for p in pulses.data:

            initial_rotation = rotation = 1

            if p.angle and cs is not None:
                if p.which is None:
                    if p.axis not in total_sigma:
                        total_sigma[p.axis] = expand(cs.sigma[p.axis], ndims - nc, shortdims)
                    rotation = center_rotation(total_sigma, p.axis, p.angle)
                else:
                    for i in p.which:
                        c = cs[i]
                        if p.axis not in separate_sigma[i]:
                            separate_sigma[i][p.axis] = expand(c.sigma[p.axis], ndims - nc, shortdims)
                        rotation = np.dot(center_rotation(separate_sigma[i], p.axis, p.angle), rotation)

            if p.bath_names is not None:
                if vectors.shape != bath.shape:
                    vectors = vectors[:bath.shape[0]]
                    # print(vectors.shape)
                properties = np.broadcast(p.bath_names, p.bath_axes, p.bath_angles)

                for name, axis, angle in properties:
                    if angle:
                        which = (bath.N == name)

                        if any(which):
                            vecs = vectors[which]
                            rotation = np.dot(bath_rotation(vecs, axis, angle), rotation)

            if initial_rotation is rotation:
                rotation = None

            p.rotation = rotation

            rots.append(rotation)

        self.delays = delays
        self.rotations = rots
        return delays, rots


_rot = {'x': 0, 'y': 1, 'z': 2}


def bath_rotation(vectors, axis, angle):
    """
    Generate rotation of the bath spins with given spin vectors.

    Args:
        vectors (ndarray with shape (n, 3, x, x)): Array of *n* bath spin vectors.
        axis (str): Axis of rotation.
        angle (float): Angle of rotation.

    Returns:
        ndarray with shape (x, x): Matrix representation of the spin rotation.

    """
    ax = _rot[axis]  # name -> index
    if (angle == np.pi) and (vectors[0][0, 0, 0] < 1):  # only works for spin-1/2
        rotation = -1j * 2 * vectors[0][ax]  # 2 here is to transform into pauli matrices
        for v in vectors[1:]:
            np.matmul(rotation, -1j * 2 * v[ax], out=rotation)
    else:
        rotation = scipy.linalg.expm(-1j * vectors[0][ax] * angle)
        for v in vectors[1:]:
            add = scipy.linalg.expm(-1j * v[ax] * angle)
            np.matmul(rotation, add, out=rotation)

    return rotation


def center_rotation(sigma, axis, angle):
    """
    Generate rotation of the central spins with given spin vectors.

    Args:
        sigma (ndarray with shape (n, 3, x, x)): Pauli matrices.
        axis (str): Axis of rotation.
        angle (float): Angle of rotation.

    Returns:
        ndarray with shape (x, x): Matrix representation of the spin rotation.

    """
    if angle == np.pi:
        rotation = -1j * sigma[axis]
    else:
        rotation = scipy.linalg.expm(-1j * sigma[axis] * angle / 2)
    return rotation



import warnings

import numpy as np
from pycce.bath.array import BathArray

from .base import set_isotopes


# hbar mu0 /4pi I have no idea about units, from mengs code
# UPDATE: derived, it checks out
# HBAR = 1.054571729
# External HF given in MHz, transform to kHz * rad
# MHZ_TO_KHZ = 2 * np.pi * 1000


def read_xyz(xyz, skiprows: int = 2, spin_types=None, isotopes=None, imap=None):
    """
    Read positions of bath spins from xyz file.

    If xyz contains names of the elements and not isotopes, changes the names to the most abundant nonzero spin
    isotopes.

    Args:
        xyz (ndarray or str):
            Either:

            - ndarray with dtype containing [('N', np.unicode_, 16), ('xyz', np.float64, (3,))], which
              contains types and positions of bath spins.
            - ndarray with coordinates of the spins.
            - Name of the xyz-file containing isotope name and xyz coordinates.

        skiprows (int): used when xyz is filename. Number of rows to skip in the file
        spin_types (SpinDict or list of tuples): SpinDict containing SpinTypes of isotopes or input to make one.
        isotopes (dict): Optional.
            Dictionary with entries: {"element" : "isotope"}, where "element" is the name of the element
            in DFT output, "isotope" is the name of the isotope.

    Returns:
        BathArray: Array of atoms read from xyz argument.

    """

    if isinstance(xyz, BathArray):
        atoms = xyz.copy()
        if spin_types is not None:
            try:
                atoms.add_type(**spin_types)
            except TypeError:
                atoms.add_type(*spin_types)
        if imap is not None:
            if atoms.imap is None:
                atoms.imap = imap
            else:
                atoms.imap = atoms.imap + imap
        return atoms

    elif isinstance(xyz, np.ndarray):
        dataset = xyz
    else:
        dt_read = np.dtype([('N', np.unicode_, 16), ('xyz', np.float64, (3,))])
        dataset = np.loadtxt(xyz, dtype=dt_read, skiprows=skiprows)

    with warnings.catch_warnings(record=True) as w:
        atoms = BathArray(array=dataset, types=spin_types, imap=imap)
    if w:
        set_isotopes(atoms, isotopes=isotopes)

    return atoms
